/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pplite-config.h"
#include "Con.hh"
#include "Gen.hh"
#include "Scalar_Prod.hh"
#include "BBox.hh"

#include <algorithm>
#include <functional>
#include <numeric>

namespace pplite {

bool
BBox::check_inv() const {
#if PPLITE_BBOX_PSEUDO_VOLUME
  if (space_dim() == 0 && !(volume.first == 0 && volume.second == Rational::one()))
    return false;
  // Note: 0-dim bbox is considered to be bounded even when is UNIVERSE.
  if (!is_bounded() && !volume.second.is_zero())
    return false;
#endif
  if (empty)
    return std::all_of(bounds.begin(), bounds.end(),
                       std::mem_fn(&Info::is_zero));
  if (!std::all_of(bounds.begin(), bounds.end(),
                   std::mem_fn(&Info::check_inv)))
    return false;
#if PPLITE_BBOX_PSEUDO_VOLUME
  dim_type num_inf = 2 * space_dim();
  for (const auto& b : bounds) {
    if (!b.inf_lb)
      --num_inf;
    if (!b.inf_ub)
      --num_inf;
  }
  if (num_inf != volume.first)
    return false;

  if (volume != compute_pseudo_volume())
    return false;
#endif
  return true;
}

bool
BBox::is_included_in(const Con& c) const {
  // If `c' is satisfied by the closest point, the whole box is.
  const auto& c_ex = c.linear_expr();
  Rational scal_prod(c.inhomo_term());
  for (dim_type i = c.space_dim(); i-- > 0; ) {
    const auto& coeff = c_ex.get(i);
    const auto s = sgn(coeff);
    if (s == 0)
      continue;
    // Check `i' infinite bounds.
    if ((s > 0 && inf_lb(i)) || (s < 0 && inf_ub(i))) {
      // Coeff should have been 0.
      return false;
    }
    if (c.is_equality() && lb(i) != ub(i))
      return false;
    // Keep building the scalar product.
    add_mul_assign(scal_prod, (s > 0) ? lb(i) : ub(i), Rational(coeff));
  }
  const auto res = sgn(scal_prod);
  return (c.is_equality()) ? (res == 0)
    : ((c.is_strict_inequality()) ? (res > 0) : (res >= 0));
}

bool
BBox::is_disjoint_from(const BBox& y) const {
  auto& x = *this;
  assert(x.space_dim() == y.space_dim());
  assert(x.check_inv());
  assert(y.check_inv());
  if (x.is_empty() || y.is_empty())
    return true;
  for (auto i = x.space_dim(); i-- > 0; )
    if ((!x.inf_lb(i) && !y.inf_ub(i) && x.lb(i) > y.ub(i))
        || (!x.inf_ub(i) && !y.inf_lb(i) && x.ub(i) < y.lb(i)))
      return true;
  return false;
}

void
BBox::glb_assign(const BBox& y) {
  auto& x = *this;
  assert(x.space_dim() == y.space_dim());
  assert(x.check_inv());
  assert(y.check_inv());
  if (x.is_empty())
    return;
  if (y.is_empty()) {
    x.set_empty();
    return;
  }

  for (auto i = x.space_dim(); i-- > 0 ; ) {
    if (!y.inf_lb(i)) {
      if (x.inf_lb(i)) {
        x.inf_lb(i) = y.inf_lb(i);
        x.lb(i) = y.lb(i);
      } else {
        x.lb(i) = std::max(x.lb(i), y.lb(i));
      }
    }
    if (!y.inf_ub(i)) {
      if (x.inf_ub(i)) {
        x.inf_ub(i) = y.inf_ub(i);
        x.ub(i) = y.ub(i);
      } else {
        x.ub(i) = std::min(x.ub(i), y.ub(i));
      }
    }
    // Check for inconsistency
    if (!x.inf_lb(i) && !x.inf_ub(i) && x.ub(i) < x.lb(i)) {
      x.set_empty();
      return;
    }
  }
  recompute_pseudo_volume();
  assert(x.check_inv());
}

void
BBox::lub_assign(const BBox& y) {
  auto& x = *this;
  assert(x.space_dim() == y.space_dim());
  assert(x.check_inv());
  assert(y.check_inv());
  if (y.is_empty())
    return;
  if (x.is_empty()) {
    x = y;
    return;
  }

  for (auto i = x.space_dim(); i-- > 0 ; ) {
    if (!x.inf_lb(i)) {
      if (y.inf_lb(i)) {
        x.inf_lb(i) = true;
        x.lb(i) = Rational(0);
      } else {
        x.lb(i) = std::min(x.lb(i), y.lb(i));
      }
    }
    if (!x.inf_ub(i)) {
      if (y.inf_ub(i)) {
        x.inf_ub(i) = true;
        x.ub(i) = Rational(0);
      } else {
        x.ub(i) = std::max(x.ub(i), y.ub(i));
      }
    }
  }
  recompute_pseudo_volume();
  assert(!x.is_empty());
  assert(x.check_inv());
}

void
BBox::time_elapse_assign(const BBox& y) {
  auto& x = *this;
  assert(x.space_dim() == y.space_dim());
  assert(x.check_inv());
  assert(y.check_inv());
  if (x.is_empty())
    return;
  if (y.is_empty()) {
    x.set_empty();
    return;
  }

  for (auto i = x.space_dim(); i-- > 0 ; ) {
    if (!x.inf_lb(i) && (y.inf_lb(i) || sgn(y.lb(i)) < 0)) {
      x.inf_lb(i) = true;
      x.lb(i) = Rational(0);
#if PPLITE_BBOX_PSEUDO_VOLUME
      ++x.volume.first;
#endif
    }
    if (!x.inf_ub(i) && (y.inf_ub(i) || sgn(y.ub(i)) > 0)) {
      x.inf_ub(i) = true;
      x.ub(i) = Rational(0);
#if PPLITE_BBOX_PSEUDO_VOLUME
      ++x.volume.first;
#endif
    }
  }
#if PPLITE_BBOX_PSEUDO_VOLUME
  if (num_rays() > 0)
    volume.second = Rational::zero();
#endif
  assert(!x.is_empty());
  assert(x.check_inv());
}

void
BBox::refine_as_integral(dim_type first, dim_type last) {
  auto& x = *this;
  assert(x.check_inv());
  assert(0 <= first && first <= last && last <= x.space_dim());
  if (x.is_empty())
    return;
  for (auto i = first; i != last; ++i) {
    if (bounds[i].refine_as_integral()) {
      x.set_empty();
      return;
    }
  }
#if PPLITE_BBOX_PSEUDO_VOLUME
  x.recompute_pseudo_volume();
#endif
  assert(x.check_inv());
}

namespace detail {

inline void
add_as_line(BBox& box, const Gen& g) {
  assert(g.space_dim() <= box.space_dim());
  assert(!box.is_empty());
  for (dim_type i = g.space_dim(); i-- > 0; ) {
    if (g.coeff(Var(i)) == 0)
      continue;
    box.unconstrain(i);
  }
}

inline void
add_as_ray(BBox& box, const Gen& g) {
  assert(g.space_dim() <= box.space_dim());
  assert(!box.is_empty());
  for (dim_type i = g.space_dim(); i-- > 0; ) {
    auto s = sgn(g.coeff(Var(i)));
    if (s < 0)
      box.unconstrain_lb(i);
    else if (s > 0)
      box.unconstrain_ub(i);
  }
}

inline void
add_as_point(BBox& box, const Gen& g) {
  assert(g.space_dim() <= box.space_dim());
  assert(!box.is_empty());
  assert(box.check_inv());

  for (dim_type i = g.space_dim(); i-- > 0; ) {
    if (box.inf_lb(i) && box.inf_ub(i))
      continue;
    Rational value(g.coeff(Var(i)), g.divisor());
    if (!box.inf_lb(i) && value < box.lb(i))
      box.lb(i) = value;
    if (!box.inf_ub(i) && value > box.ub(i))
      box.ub(i) = value;
  }
}

inline void
init_with_point(BBox& box, const Gen& g) {
  assert(g.space_dim() <= box.space_dim());
  assert(g.is_point());
  assert(box.is_empty());
  assert(box.check_inv());

  for (dim_type i = g.space_dim(); i-- > 0; ) {
    assert(!box.inf_lb(i) && !box.inf_ub(i));
    Rational value(g.coeff(Var(i)), g.divisor());
    box.lb(i) = value;
    box.ub(i) = value;
  }

  box.empty = false;
  box.volume.first = 0;
  box.volume.second = Rational::one();
}

} // namespace detail

void
BBox::add_gen(const Gen& g) {
  assert(g.space_dim() <= space_dim());
  assert(!(space_dim() == 0 && g.is_line_or_ray()));

  if (space_dim() == 0) {
    if (is_empty()) {
      assert(g.is_point());
      empty = false;
      volume.second = Rational::one();
    }
    assert(check_inv());
    return;
  }

  auto& x = *this;
  if (is_empty()) {
    detail::init_with_point(x, g);
    assert(!is_empty());
    assert(check_inv());
    return;
  }

  auto type = g.type();
  switch (type) {
  case Gen::LINE:
    detail::add_as_line(x, g);
    break;
  case Gen::RAY:
    detail::add_as_ray(x, g);
    break;
  case Gen::POINT:
  case Gen::CLOSURE_POINT:
  default:
    assert(type == Gen::POINT || type == Gen::CLOSURE_POINT);
    detail::add_as_point(x, g);
    break;
  }
  recompute_pseudo_volume();
  assert(check_inv());
}

template <typename Iter>
void
BBox::add_gens(Iter first, Iter last) {
  if (first == last)
    return;
  if (is_empty()) {
    auto iter = std::find_if(first, last,
                             std::mem_fn(&Gen::is_point));
    assert(iter != last);
    add_gen(*iter);
    add_gens(first, iter);
    ++iter;
    add_gens(iter, last);
    return;
  }

  assert(!is_empty());
  while (first != last) {
    add_gen(*first);
    ++first;
  }
  assert(check_inv());
}

void
BBox::add_gens(const Gens& gs) {
  add_gens(gs.cbegin(), gs.cend());
}

void
BBox::add_gens_as_rays(const Gens& gs) {
  assert(!is_empty());
  for (const auto& g : gs)
    detail::add_as_ray(*this, g);
  assert(check_inv());
}

void
BBox::affine_image(Var var, Linear_Expr expr,
                   Integer inhomo, Integer den) {
  assert(space_dim() > 0);
  assert(var.space_dim() <= space_dim());
  assert(expr.space_dim() <= space_dim());
  assert(den != 0);

  if (is_empty())
    return;
  if (den < 0) {
    neg_assign(expr);
    neg_assign(inhomo);
    neg_assign(den);
  }
  assert(den > 0);

  auto vi = var.id();

  Index_Set pos, neg;
  auto get_sign_vars = [&expr, &pos, &neg]() {
    for (auto i = expr.space_dim(); i-- > 0; ) {
      auto s = sgn(expr.get(i));
      if (s == 0)
        continue;
      (s > 0) ?  pos.set(i) : neg.set(i);
    }
  };

  auto compute_new_bound =
    [this, &expr, &inhomo, &den](const Index_Set& lbs,
                                 const Index_Set& ubs) {
    Rational val(inhomo);
    for (auto i : lbs)
      add_mul_assign(val, Rational(expr[i]), lb(i));
    for (auto i : ubs)
      add_mul_assign(val, Rational(expr[i]), ub(i));
    val = val / Rational(den);
    return val;
  };

  auto is_inf_lb = [this](dim_type i) { return inf_lb(i); };
  auto is_inf_ub = [this](dim_type i) { return inf_ub(i); };
  bool becomes_inf_lb = false;
  bool becomes_inf_ub = false;

  get_sign_vars();

  if (std::any_of(pos.begin(), pos.end(), is_inf_lb)
      || std::any_of(neg.begin(), neg.end(), is_inf_ub)) {
    becomes_inf_lb = true;
    unconstrain_lb(vi);
  }
  if (std::any_of(pos.begin(), pos.end(), is_inf_ub)
      || std::any_of(neg.begin(), neg.end(), is_inf_lb)) {
    becomes_inf_ub = true;
    unconstrain_ub(vi);
  }

  if (!becomes_inf_lb) {
    inf_lb(vi) = false;
    lb(vi) = compute_new_bound(pos, neg);
  }

  if (!becomes_inf_ub) {
    inf_ub(vi) = false;
    ub(vi) = compute_new_bound(neg, pos);
  }

  recompute_pseudo_volume();
  assert(check_inv());
}

void
BBox::ascii_dump(std::ostream& os) const {
  os << "empty = " << empty << ", ";
#if PPLITE_BBOX_PSEUDO_VOLUME
  os << "num_inf = " << volume.first << ", ";
  os << "pseudo_vol = " << volume.second << ", ";
#endif
  for (dim_type i = 0; i != num_rows(bounds); ++i) {
    if (i > 0)
      os << ", ";
    Var(i).print(os);
    if (inf_lb(i))
      os << " in (-inf";
    else
      os << " in [" << lb(i);
    if (inf_ub(i))
      os << ", +inf)";
    else
      os << ", " << ub(i) << "]";
  }
  os << "\n";
}

} // namespace pplite
