/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pplite-config.h"
#include "Bit_Row.hh"

#include <algorithm>
#include <cassert>
#include <iostream>

namespace pplite {

dim_type
Bit_Row::first() const {
  const auto vec_size = vec->_mp_size;
  assert(vec_size >= 0);
  mp_srcptr p = vec->_mp_d;
  for (mp_size_t li = 0; li < vec_size; ++li, ++p) {
    const mp_limb_t limb = *p;
    if (limb != 0)
      return li * bits_per_limb + first_one(limb);
  }
  return end_pos;
}

dim_type
Bit_Row::next(dim_type position) const {
  ++position;
  auto li = position / bits_per_limb;
  const auto vec_size = vec->_mp_size;
  assert(vec_size >= 0);
  if (li >= vec_size)
    return end_pos;

  // Get the first limb.
  mp_srcptr p = vec->_mp_d + li;
  // Mask off any bits before `position' in the first limb.
  auto limb = *p & (limb_all_set << (position % bits_per_limb));

  while (true) {
    if (limb != 0)
      return li * bits_per_limb + first_one(limb);
    ++li;
    if (li == vec_size)
      break;
    ++p;
    limb = *p;
  }
  return end_pos;
}

dim_type
Bit_Row::last() const {
  auto li = vec->_mp_size;
  assert(li >= 0);
  if (li == 0)
    return end_pos;
  --li;
  const mp_srcptr p = vec->_mp_d + li;
  const auto limb = *p;
  assert(limb != 0);
  return li * bits_per_limb + last_one(limb);
}

dim_type
Bit_Row::prev(dim_type position) const {
  if (position == 0)
    return end_pos;

  --position;
  const auto vec_size = vec->_mp_size;
  assert(vec_size > 0);
  auto li = position / bits_per_limb;

  mp_limb_t limb;
  mp_srcptr p = vec->_mp_d;

  // Get the first limb.
  if (li >= vec_size) {
    li = vec_size - 1;
    p += li;
    limb = *p;
  } else {
    const mp_limb_t mask
      = limb_all_set
      >> (bits_per_limb - 1U - position % bits_per_limb);
    p += li;
    limb = *p & mask;
  }

  while (true) {
    if (limb != 0)
      return li * bits_per_limb + last_one(limb);
    if (li == 0)
      break;
    --li;
    --p;
    limb = *p;
  }
  return end_pos;
}

void
Bit_Row::set_until(dim_type k) {
  assert(k >= 0);
  if (k == 0)
    return;
  // This resizes vec (if needed).
  mpz_setbit(vec, k - 1);
  const auto k_div = k / bits_per_limb;
  for (auto i = 0; i < k_div; ++i)
    vec->_mp_d[i] = limb_all_set;
  const auto k_mod = k % bits_per_limb;
  if (k_mod > 0)
    vec->_mp_d[k_div] |= (limb_all_set >> (bits_per_limb - k_mod));
}

void
Bit_Row::complement_until(const dim_type k) {
  assert(k >= 0);
  if (k == 0)
    return;
  auto k_div = k / bits_per_limb;
  auto k_mod = k % bits_per_limb;
  auto k_size = k_div + (k_mod == 0 ? 0 : 1);
  assert(k_size > 0);
  auto old_size = vec->_mp_size;

  if (old_size < k_size) {
    // Have to increase size.
    mpz_realloc2(vec, k_size * bits_per_limb);
    vec->_mp_size = k_size;
    mp_ptr p = vec->_mp_d + k_size;
    // Maybe complement last (incomplete) limb.
    if (k_mod > 0) {
      --p;
      *p = (limb_all_set >> (bits_per_limb - k_mod));
    }
    // Complement new (complete) limbs.
    for (auto i = k_div - old_size; i-- > 0; ) {
      --p;
      *p = limb_all_set;
    }
    // Complement old limbs.
    for (auto i = old_size; i-- > 0; ) {
      --p;
      *p = ~(*p);
    }
    assert(check_inv());
    return;
  }

  // No need to increase size.
  assert(old_size >= k_size);
  bool check_for_zero = (old_size == k_size);
  mp_ptr p = vec->_mp_d + k_size;

  if (k_mod > 0) {
    // Partial complement for this limb.
    --p;
    mp_limb_t mask = (limb_all_set >> (bits_per_limb - k_mod));
    mp_limb_t p_lower = (~(*p) & mask);
    mp_limb_t p_upper = (*p & ~mask);
    *p = (p_lower | p_upper);
    if (check_for_zero && (*p == 0))
      --(vec->_mp_size);
    else
      check_for_zero = false;
  }
  // Full complement for previous limbs.
  for (auto i = k_div; i-- > 0; ) {
    --p;
    *p = ~(*p);
    if (check_for_zero && (*p == 0))
      --(vec->_mp_size);
    else
      check_for_zero = false;
  }
  assert(check_inv());
}

void
Bit_Row::print_bits(mp_limb_t v) {
  Bit_Row row;
  mpz_set_ui(row.vec, v);
  row.ascii_dump(std::cerr);
}

void
Bit_Row::copy_from(dim_type dst_pos,
                   const Bit_Row& src, dim_type src_pos, dim_type count) {
  auto dst_div = dst_pos / bits_per_limb;
  auto dst_mod = dst_pos % bits_per_limb;
  auto src_div = src_pos / bits_per_limb;
  auto src_mod = src_pos % bits_per_limb;

#ifndef NDEBUG
  auto min_bits = dst_pos + count;
  auto min_limbs = min_bits / bits_per_limb;
  if (min_bits % bits_per_limb != 0)
    ++min_limbs;
  assert(vec->_mp_size >= min_limbs);
#endif // NDEBUG

  mp_ptr dst_ptr = vec->_mp_d + dst_div;
  mp_srcptr src_ptr = src.vec->_mp_d + src_div;

  // First (partial) limb.
  if (dst_mod < src_mod) {
    auto partial_count = std::min(bits_per_limb - src_mod, count);
    *dst_ptr = limb_partial_copy(*src_ptr, src_mod,
                                 *dst_ptr, dst_mod, partial_count);
    if (count <= partial_count) {
      assert(check_inv());
      return;
    }
    count -= partial_count;
    ++src_ptr;
    assert(src_mod + partial_count == bits_per_limb);
    src_mod = 0;
    dst_mod += partial_count;
    partial_count = std::min(bits_per_limb - dst_mod, count);
    *dst_ptr = limb_partial_copy(*src_ptr, src_mod,
                                 *dst_ptr, dst_mod, partial_count);
    if (count <= partial_count) {
      assert(check_inv());
      return;
    }
    count -= partial_count;
    ++dst_ptr;
    src_mod = partial_count;
    dst_mod = 0;
  } else {
    assert(dst_mod >= src_mod);
    auto partial_count = std::min(bits_per_limb - dst_mod, count);
    *dst_ptr = limb_partial_copy(*src_ptr, src_mod,
                                 *dst_ptr, dst_mod, partial_count);
    if (count <= partial_count) {
      assert(check_inv());
      return;
    }
    count -= partial_count;
    ++dst_ptr;
    if (dst_mod == src_mod) {
      ++src_ptr;
      src_mod = 0;
      dst_mod = 0;
    } else {
      src_mod += partial_count;
      dst_mod = 0;
    }
  }
  // Now dst limb is aligned.
  assert(dst_mod == 0);

  // Copy other (complete) limbs.
  auto count_div = count / bits_per_limb;
  if (src_mod == 0) {
    // Lucky case: also src is aligned.
    while (count_div > 0) {
      *dst_ptr = *src_ptr;
      ++src_ptr;
      ++dst_ptr;
      --count_div;
    }
    // Last (partial) limb.
    count %= bits_per_limb;
    if (count > 0)
      *dst_ptr = limb_partial_copy(*src_ptr, 0, *dst_ptr, 0, count);
  } else {
    assert(src_mod > 0);
    // src is misaligned: perform shifted copies.
    while (count_div > 0) {
      *dst_ptr = limb_shifted_copy(*src_ptr, *(src_ptr+1), src_mod);
      ++src_ptr;
      ++dst_ptr;
      --count_div;
    }
    // Last limb.
    count %= bits_per_limb;
    if (count > 0) {
      auto partial_count = std::min(bits_per_limb - src_mod, count);
      *dst_ptr = limb_partial_copy(*src_ptr, src_mod,
                                   *dst_ptr, 0, partial_count);
      if (count <= partial_count) {
        assert(check_inv());
        return;
      }
      count -= partial_count;
      ++src_ptr;
      assert(src_mod + partial_count == bits_per_limb);
      src_mod = 0;
      dst_mod += partial_count;
      assert(count <= bits_per_limb - dst_mod);
      *dst_ptr = limb_partial_copy(*src_ptr, src_mod,
                                   *dst_ptr, dst_mod, count);
    }
  }
  assert(check_inv());
}

int
compare(const Bit_Row& x, const Bit_Row& y) {
  const auto& x_vec = x.impl();
  const auto& y_vec = y.impl();
  const auto x_size = x_vec->_mp_size;
  const auto y_size = y_vec->_mp_size;
  assert(x_size >= 0 && y_size >= 0);
  auto size = std::min(x_size, y_size);
  mp_srcptr xp = x_vec->_mp_d;
  mp_srcptr yp = y_vec->_mp_d;
  while (size > 0) {
    const mp_limb_t xl = *xp;
    const mp_limb_t yl = *yp;
    if (xl != yl) {
      // Get the ones where they are different.
      const mp_limb_t diff = xl ^ yl;
      // First bit that is different.
      const mp_limb_t mask = diff & ~(diff-1);
      return ((xl & mask) != 0) ? 1 : -1;
    }
    ++xp;
    ++yp;
    --size;
  }
  return (x_size == y_size) ? 0 : ((x_size > y_size) ? 1 : -1);
}

bool
subset_or_equal(const Bit_Row& x, const Bit_Row& y) {
#if PPLITE_LOW_LEVEL_COUNTERS
  LLOp_Stats::bump_sat_incl();
#endif
  const auto& x_vec = x.impl();
  const auto& y_vec = y.impl();
  auto x_size = x_vec->_mp_size;
  const auto y_size = y_vec->_mp_size;
  assert(x_size >= 0 && y_size >= 0);
  if (x_size > y_size)
    return false;
  mp_srcptr xp = x_vec->_mp_d;
  mp_srcptr yp = y_vec->_mp_d;
  while (x_size > 0) {
    if ((*xp & ~*yp) != 0)
      return false;
    ++xp;
    ++yp;
    --x_size;
  }
  return true;
}

bool
subset_or_equal(const Bit_Row& x, const Bit_Row& y,
                bool& strict_subset) {
#if PPLITE_LOW_LEVEL_COUNTERS
  LLOp_Stats::bump_sat_incl();
#endif
  const auto& x_vec = x.impl();
  const auto& y_vec = y.impl();
  auto x_size = x_vec->_mp_size;
  const auto y_size = y_vec->_mp_size;
  assert(x_size >= 0 && y_size >= 0);
  if (x_size > y_size)
    return false;
  mp_srcptr xp = x_vec->_mp_d;
  mp_srcptr yp = y_vec->_mp_d;
  strict_subset = (x_size < y_size);
  mp_limb_t xl;
  mp_limb_t yl;
  if (strict_subset) {
    while (x_size > 0) {
      xl = *xp;
      yl = *yp;
      if ((xl & ~yl) != 0) {
        return false;
      }
    strict_subset_next:
      ++xp;
      ++yp;
      --x_size;
    }
  }
  else {
    while (x_size > 0) {
      xl = *xp;
      yl = *yp;
      if (xl != yl) {
        if ((xl & ~yl) != 0) {
          return false;
        }
        strict_subset = true;
        goto strict_subset_next;
      }
      ++xp;
      ++yp;
      --x_size;
    }
  }
  return true;
}

bool
strict_subset(const Bit_Row& x, const Bit_Row& y) {
#if PPLITE_LOW_LEVEL_COUNTERS
  LLOp_Stats::bump_sat_incl();
#endif
  const auto& x_vec = x.impl();
  const auto& y_vec = y.impl();
  auto x_size = x_vec->_mp_size;
  const auto y_size = y_vec->_mp_size;
  assert(x_size >= 0 && y_size >= 0);
  if (x_size > y_size)
    return false;
  bool different = (x_size < y_size);
  mp_srcptr xp = x_vec->_mp_d;
  mp_srcptr yp = y_vec->_mp_d;
  while (x_size > 0) {
    const mp_limb_t xl = *xp;
    const mp_limb_t yl = *yp;
    if ((xl & ~yl) != 0) {
      return false;
    }
    if (!different && xl != yl) {
      different = true;
    }
    ++xp;
    ++yp;
    --x_size;
  }
  return different;
}

bool
Bit_Row::intersects(const Bit_Row& y) const {
#if PPLITE_LOW_LEVEL_COUNTERS
  LLOp_Stats::bump_sat_count();
#endif
  const Bit_Row& x = *this;
  auto min_size = std::min(x.vec->_mp_size, y.vec->_mp_size);
  assert(min_size >= 0);
  mp_srcptr xp = x.vec->_mp_d;
  mp_srcptr yp = y.vec->_mp_d;
  while (min_size > 0) {
    if ((*xp & *yp) != 0) {
      return true;
    }
    ++xp;
    ++yp;
    --min_size;
  }
  return false;
}

bool
Bit_Row::check_inv() const {
  const auto vec_size = vec->_mp_size;
  const auto vec_alloc = vec->_mp_alloc;
  return vec_size >= 0
    && vec_alloc >= vec_size
    && (vec_size == 0 || mpz_getlimbn(vec, vec_size - 1) != 0);
}

void
Bit_Row::union_helper(const Bit_Row& y, const Bit_Row& z) {
#if PPLITE_LOW_LEVEL_COUNTERS
  LLOp_Stats::bump_sat_union();
#endif
  auto y_size = y.vec->_mp_size;
  auto z_size = z.vec->_mp_size;
  assert(y_size <= z_size);
  assert(vec->_mp_alloc >= z_size);
  vec->_mp_size = z.vec->_mp_size;
  mp_srcptr yp = y.vec->_mp_d;
  mp_srcptr zp = z.vec->_mp_d;
  mp_ptr p = vec->_mp_d;
  z_size -= y_size;
  while (y_size > 0) {
    *p = *yp | * zp;
    ++yp;
    ++zp;
    ++p;
    --y_size;
  }
  while (z_size > 0) {
    *p = *zp;
    ++zp;
    ++p;
    --z_size;
  }
}

dim_type
Bit_Row::count_ones_in_union(const Bit_Row& y) const {
#if PPLITE_LOW_LEVEL_COUNTERS
  LLOp_Stats::bump_sat_count();
#endif
  const Bit_Row& x = *this;
  mp_srcptr xp = x.vec->_mp_d;
  mp_srcptr yp = y.vec->_mp_d;
  const auto x_size = x.vec->_mp_size;
  const auto y_size = y.vec->_mp_size;
  const auto sz = std::min(x_size, y_size);
  dim_type res = 0;
  mp_limb_t limb;
  for (mp_size_t i = 0; i < sz; ++i, ++xp, ++yp) {
    limb = *xp | *yp;
    res += mpn_popcount(&limb, 1);
  }
  if (sz < x_size)
    res += mpn_popcount(xp, x_size - sz);
  else if (sz < y_size)
    res += mpn_popcount(yp, y_size - sz);
  return res;
}

void
Bit_Row::ascii_dump(std::ostream& s) const {
  const Bit_Row& x = *this;
  s << x[0];
  if (!x.empty()) {
    const char separator = ' ';
    for (dim_type i = 1, i_end = x.last(); i <= i_end; ++i)
      s << separator << x[i];
  }
  s << "\n";
}

} // namespace pplite
