/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pplite-config.h"

#include "F_Poly.hh"

#include <algorithm>
#include <map>
#include <memory>

namespace pplite {

F_Poly::F_Poly(const Poly& ph)
  : F_Poly(ph.space_dim(), ph.topology()) {
  const auto& cs = ph.cons();
  add_cons(cs.begin(), cs.end());
}

namespace detail {

using Block = F_Poly::Block;
using Blocks = F_Poly::Blocks;
using Factor = F_Poly::Factor;
using Factors = F_Poly::Factors;

template <typename T>
inline void
splice(std::vector<T>& xs, std::vector<T>& ys) {
  xs.reserve(xs.size() + ys.size());
  for (auto& y : ys)
    xs.push_back(std::move(y));
}

template <typename T, typename U, typename Pred>
void
synchronous_sort(std::vector<T>& xs, std::vector<U>& ys,
                 Pred x_lesser) {
  assert(num_rows(xs) == num_rows(ys));

  // Compute the order of the elements according to xs elems.
  auto pred = [&xs, &x_lesser] (dim_type i, dim_type j)
              { return x_lesser(xs[i], xs[j]); };
  std::list<dim_type> order;
  for (auto i = num_rows(xs); i-- > 0; )
    order.insert(std::lower_bound(order.begin(), order.end(), i, pred), i);

  assert(num_rows(xs) == num_rows(order));

  // Sort xs and ys according to the computed order.
  std::vector<T> xs_res;
  xs_res.reserve(xs.size());
  std::vector<U> ys_res;
  ys_res.reserve(ys.size());
  for (auto i : order) {
    xs_res.push_back(std::move(xs[i]));
    ys_res.push_back(std::move(ys[i]));
  }

  xs = std::move(xs_res);
  ys = std::move(ys_res);
}

inline void
add_var(Block& b, Var v) {
  if (std::find(b.begin(), b.end(), v.id()) == b.end())
    b.push_back(v.id());
}

inline Block
extract_block(const Linear_Expr& e) {
  Block var_block;
  for (dim_type i = 0; i != e.space_dim(); ++i)
    if (e.get(Var(i)) != 0)
      var_block.push_back(i);
  return var_block;
}

inline Block
extract_block(const Con& c) {
  return extract_block(c.linear_expr());
}

inline Block
extract_block(const Gen& g) {
  assert(g.is_ray() || g.is_line());
  return extract_block(g.linear_expr());
}

inline bool
are_disjoint(const Block& b1, const Block& b2) {
  return std::find_first_of(b1.begin(), b1.end(),
                            b2.begin(), b2.end()) == b1.end();
}

template <typename T>
dim_type
count_intersections(const Blocks& bs, const T& x) {
  Block b = extract_block(x);
  auto intersects = [&b] (const Block& b1) { return not are_disjoint(b1, b); };
  return std::count_if(bs.begin(), bs.end(), intersects);
}

inline Var
convert(Var v, const Block& b) {
  for (dim_type i = num_rows(b); i-- > 0; ) {
    if (b[i] == v.id())
      return Var(i);
  }
  assert(false);
  abort();
}

inline Linear_Expr
convert(const Linear_Expr& expr, const Block& b) {
  Linear_Expr res;
  for (dim_type i = num_rows(b); i-- > 0; )
    add_mul_assign(res, expr.get(b[i]), Var(i));
  return res;
}

inline Con
convert(const Con& c, const Block& b) {
  return Con(convert(c.linear_expr(), b), c.inhomo_term(), c.type());
}

inline Gen
convert(const Gen& g, const Block& b) {
  return Gen(g.type(), convert(g.linear_expr(), b), g.impl().inhomo);
}

inline Block
convert_back(const Block& int_b, const Block& ext_b) {
  auto res = int_b;
  for (auto& i : res)
    i = ext_b[i];
  return res;
}

inline Linear_Expr
convert_back(const Linear_Expr& expr, const Block& b) {
  Linear_Expr res;
  for (dim_type i = num_rows(b); i-- > 0; )
    add_mul_assign(res, expr.get(i), Var(b[i]));
  return res;
}

inline Con
convert_back(const Con& c, const Block& b) {
  return Con(convert_back(c.linear_expr(), b), c.inhomo_term(), c.type());
}

inline Gen
convert_back(const Gen& g, const Block& b) {
  return Gen(g.type(), convert_back(g.linear_expr(), b), g.impl().inhomo);
}

inline bool
contains(const Block& b, dim_type i) {
  return std::find(b.begin(), b.end(), i) != b.end();
}

inline bool
contains(const Block& b1, const Block& b2) {
  return std::all_of(b2.begin(), b2.end(),
                     [&b1](dim_type i) { return contains(b1, i); });
}

inline dim_type
has_block_index(const Blocks& bs, Var v) {
  for (dim_type i = num_rows(bs); i-- > 0; ) {
    if (contains(bs[i], v.id()))
      return i;
  }
  return not_a_dim();
}

inline dim_type
get_block_index(const Blocks& bs, Var v) {
  auto res = has_block_index(bs, v);
  assert(res != not_a_dim());
  return res;
}

inline dim_type
get_block_index(const Blocks& bs, const Linear_Expr& expr) {
  // Note: assumes that bs contains a block b containing all dims
  // having a non-zero coeff in expr.
  auto i = expr.space_dim();
  do { --i; } while (expr[i].is_zero());
  return get_block_index(bs, Var(i));
}

inline dim_type
get_block_index(const Blocks& bs, const Con& c) {
  return get_block_index(bs, c.linear_expr());
}

inline dim_type
get_containing_block_index(const Blocks& bs, const Block& b) {
  for (dim_type i = num_rows(bs); i-- > 0; ) {
    if (!are_disjoint(bs[i], b)) {
      assert(contains(bs[i], b));
      return i;
    }
  }
  assert(false);
  abort();
}

inline dim_type
has_containing_block_index(const Blocks& bs, const Block& b) {
  for (dim_type i = num_rows(bs); i-- > 0; ) {
    if (are_disjoint(bs[i], b))
      continue;
    return contains(bs[i], b) ? i : not_a_dim();
  }
  return not_a_dim();
}

inline void
reduce_blocks(Blocks& bs, Var v) {
  for (auto& b : bs)
    for (auto& i : b)
      if (i > v.id())
        --i;
}

inline void
concat(Block& b1, const Block& b2) {
  b1.insert(b1.end(), b2.begin(), b2.end());
}

inline void
concat(Block& b1, Factor& f1, const Block& b2, const Factor& f2) {
  concat(b1, b2);
  f1.concatenate_assign(f2);
}

Index_Set
joining_blocks(const Blocks& bs, const Block& joiner) {
  Index_Set joining;
  for (auto i = num_rows(bs); i-- > 0; ) {
    if (!are_disjoint(bs[i], joiner))
      joining.set(i);
  }
  return joining;
}

void
join_blocks(Blocks& bs, const Block& joiner) {
  auto joining = joining_blocks(bs, joiner);
  assert(joining.size() > 0);
  if (joining.size() == 1)
    return;
  auto join_idx = joining.first();
  joining.clear(join_idx);
  for (auto i : joining)
    concat(bs[join_idx], bs[i]);
  erase_using_sorted_indices(bs, joining);
}

std::pair<Block, Factor>
join_factors(const Blocks& bs, const Factors& fs,
             const Index_Set& joining) {
  auto i = joining.first();
  auto res = std::make_pair(bs[i], fs[i]);
  for (auto j : joining) {
    if (i == j) continue;
    concat(res.first, bs[j]);
    res.second.concatenate_assign(fs[j]);
  }
  return res;
}

void
remap_factor(Factor& f, const Block& src_b, const Block& dst_b) {
  const auto sz = num_rows(src_b);
  assert(sz == num_rows(dst_b));
  Dims pf(sz);
  for (auto j = 0; j != sz; ++j) {
    for (auto k = 0; k != sz; ++k) {
      if (src_b[j] == dst_b[k]) {
        pf[j] = k;
        break;
      }
    }
  }
  f.map_space_dims(pf);
}

Factor
join_and_remap(const Blocks& bs, const Factors& fs,
               const Index_Set& joining, const Block& dst_b) {
  Block src_b;
  Factor src_f;
  std::tie(src_b, src_f) = join_factors(bs, fs, joining);
  if (src_b != dst_b)
    remap_factor(src_f, src_b, dst_b);
  return src_f;
}

Blocks
block_partition(const Factor& f) {
  assert(!f.is_empty() && f.space_dim() > 0);
  const auto sdim = f.space_dim();
  Blocks res(sdim);
  for (auto i = 0; i != sdim; ++i)
    res[i].push_back(i);
  if (f.space_dim() == 1)
    return res;
  for (const auto& c : f.normalized_cons()) {
    Block b_c = extract_block(c);
    join_blocks(res, b_c);
    if (num_rows(res) == 1)
      return res;
  }
  return res;
}

inline bool
can_be_factored(const Factor& f) {
  return !f.is_empty()
    && (f.space_dim() > 1)
    && (block_partition(f).size() > 1);
}

Factors
project_on_partition(const Factor& f, const Blocks& bs) {
  Factors res;
  const auto nbs = num_rows(bs);
  res.reserve(nbs);
  // init with universe of correct dim and topology
  for (const auto& b : bs)
    res.emplace_back(num_rows(b), f.topology());
  // partition (and convert) the constraints
  for (const auto& c : f.cons()) {
    auto i = get_block_index(bs, c);
    assert(0 <= i && i < nbs);
    res[i].add_con(convert(c, bs[i]));
  }
  // minimize // FIXME: needed? worth?
  for (auto& ri : res)
    ri.minimize();
  return res;
}

std::pair<Blocks, Factors>
factorize(Block& b, Factor& f) {
  // static Local_Stats stats("detail::factorize");
  // Local_Clock clock(stats);
  assert(!f.is_empty() && f.space_dim() > 0);
  // check for non-trivial block partitioning
  std::pair<Blocks, Factors> res;
  Blocks int_bs = block_partition(f);
  const auto num_proj = num_rows(int_bs);
  if (num_proj == 1)
    return res;
  // actually compute partition projections
  Factors proj_fs = project_on_partition(f, int_bs);
  // add projected blocks and factors to result, starting from second
  for (auto i = 1; i != num_proj; ++i) {
    res.first.push_back(convert_back(int_bs[i], b));
    res.second.push_back(std::move(proj_fs[i]));
  }
  // assign to b and f the first projection
  b = convert_back(int_bs[0], b);
  f = std::move(proj_fs[0]);
  return res;
}

inline Block
block_union(const Block& b1, const Block& b2) {
  Block res = b1;
  for (auto i2 : b2) {
    if (!contains(b1, i2))
      res.push_back(i2);
  }
  return res;
}

Blocks
least_upper_bound(const Blocks& bs1, const Blocks& bs2) {
  // static Local_Stats stats("blocks lub");
  // Local_Clock clock(stats);
  Blocks lub;
  for (const auto& b1 : bs1) {
    bool first = true;
    for (const auto& b2 : bs2) {
      if (are_disjoint(b1, b2))
        continue;
      if (first) {
        lub.push_back(block_union(b1, b2));
        first = false;
      } else
        lub.back() = block_union(lub.back(), b2);
    }
  }

  Index_Set to_be_rem;
  const auto nb = num_rows(lub);
  for (auto i = 0; i != nb; ++i) {
    if (to_be_rem.test(i))
      continue;
    auto j1 = i;
    for (auto j2 = j1 + 1; j2 != nb; ++j2) {
      if (to_be_rem.test(j2))
        continue;
      if (are_disjoint(lub[j1], lub[j2]))
        continue;
      lub[j2] = block_union(lub[j1], lub[j2]);
      to_be_rem.set(j1);
      j1 = j2;
    }
  }
  erase_using_sorted_indices(lub, to_be_rem);
  return lub;
}

struct Refactor_Proxy {
  const Blocks& bs_;
  const Factors& fs_;
  const Blocks& lub_;
  mutable std::vector<std::unique_ptr<Factor>> cache;

  const Factor& operator[](dim_type i) const {
    assert(0 <= i && i <= num_rows(lub_));
    if (cache[i])
      return *cache[i];
    Index_Set joiner = joining_blocks(bs_, lub_[i]);
    if (joiner.size() == 1 && bs_[joiner.first()] == lub_[i])
      // No need to join&remap
      return fs_[joiner.first()];
    else {
      cache[i].reset(new Factor(join_and_remap(bs_, fs_, joiner, lub_[i])));
      return *cache[i];
    }
  }

  Refactor_Proxy(const Blocks& bs, const Factors& fs, const Blocks& lub)
    : bs_(bs), fs_(fs), lub_(lub), cache(num_rows(lub_)) {
    assert(num_rows(bs_) == num_rows(fs_));
    assert(num_rows(bs_) >= num_rows(lub_));
  }
  ~Refactor_Proxy() = default;

  Refactor_Proxy(const Refactor_Proxy&) = delete;
  Refactor_Proxy& operator=(const Refactor_Proxy&) = delete;
  Refactor_Proxy(Refactor_Proxy&&) = delete;
  Refactor_Proxy& operator=(Refactor_Proxy&&) = delete;
};

} // namespace detail

void
F_Poly::factorize() const {
  minimize();
  if (is_empty() || space_dim() == 0)
    return;
#if 0
  struct Counter {
    long calls = 0;
    long non_fact = 0;
    long fact = 0;
    long blocks_in = 0;
    long blocks_out = 0;

    void check(const Factors& fs) {
      ++calls;
      auto in = 0;
      auto out = 0;
      for (const auto& f : fs) {
        Blocks new_bs = detail::block_partition(f);
        if (new_bs.size() > 1) {
          ++in;
          out += new_bs.size();
          std::vector<dim_type> vs;
          for (const auto& b : new_bs)
            vs.push_back(b.size());
          sort(vs.begin(), vs.end());
          std::cerr << "=== fact: " << f.space_dim()
                    << " -> ";
          for (auto s : vs) { std::cerr << s << " , "; }
          std::cerr << "\n";
        }
      }
      if (in == 0)
        ++non_fact;
      else {
        ++fact;
        blocks_in += in;
        blocks_out += out;
      }
    }

    ~Counter() {
      std::cerr << "\n=== factorize counter:\n";
      std::cerr << "=== num_calls = " << calls << "\n";
      std::cerr << "=== non factorizable = " << non_fact << "\n";
      std::cerr << "=== factorizable = " << fact << "\n";
      std::cerr << "=== blocks factored = " << blocks_in << "\n";
      std::cerr << "=== blocks out = " << blocks_out << "\n";
      std::cerr << "=== avg blocks out = "
                << ((double) blocks_out) / blocks_in << "\n";
    }
  };

  static Counter cc;
  cc.check(factors);

#else
  using namespace detail;

  // logically const
  auto& x = const_cast<F_Poly&>(*this);

  Blocks new_bs;
  Factors new_fs;
  const auto nb = num_rows(x.blocks);
  for (auto i = 0; i != nb; ++i) {
    std::tie(new_bs, new_fs) = detail::factorize(x.blocks[i], x.factors[i]);
    // Note: fine, since nb is precomputed.
    splice(x.blocks, new_bs);
    splice(x.factors, new_fs);
  }
  assert(x.check_inv());
#endif
}

void
F_Poly::m_swap(F_Poly& y) noexcept {
  using std::swap;
  swap(dim, y.dim);
  swap(empty, y.empty);
  swap(topol, y.topol);
  swap(blocks, y.blocks);
  swap(factors, y.factors);
}

dim_type
F_Poly::merge(const Block& b) {
  // CHECKME: may be too preemptive, but covers the case in which an equality
  // of size_dim() == 1, breaks the normalization.
  is_normalized = false;

  // static Local_Stats stats("merge");
  // Local_Clock clock(stats);
  assert(b.size() > 0);
  if (b.size() == 1) {
    Var v(b.front());
    return detail::get_block_index(blocks, v);
  }
  auto idx = detail::has_containing_block_index(blocks, b);
  if (idx != not_a_dim())
    return idx;

  const auto nb = num_rows(blocks);
  dim_type merge_idx = not_a_dim();
  Blocks merge_bs;
  merge_bs.reserve(nb);
  Factors merge_fs;
  merge_fs.reserve(nb);
  for (auto i = 0; i != nb; ++i) {
    auto& bi = blocks[i];
    auto& fi = factors[i];
    if (detail::are_disjoint(bi, b)) {
      merge_bs.push_back(std::move(bi));
      merge_fs.push_back(std::move(fi));
    } else if (merge_idx == not_a_dim()) {
      assert(i == num_rows(merge_bs));
      merge_idx = i;
      merge_bs.push_back(std::move(bi));
      merge_fs.push_back(std::move(fi));
    } else
      detail::concat(merge_bs[merge_idx], merge_fs[merge_idx], bi, fi);
  }
  assert(merge_idx != not_a_dim());
  blocks = std::move(merge_bs);
  factors = std::move(merge_fs);

  return merge_idx;
}

void
F_Poly::sync(const Blocks& bs) const {
  auto& x = const_cast<F_Poly&>(*this);
  const auto nb = num_rows(bs);
  assert(num_rows(x.blocks) >= nb);
  for (auto i = 0; i != nb; ++i) {
    auto j = x.merge(bs[i]);
    assert(j >= i);
    if (x.blocks[j] != bs[i]) {
      detail::remap_factor(x.factors[j], x.blocks[j], bs[i]);
      x.blocks[j] = bs[i];
    }
    if (i != j) {
      using std::swap;
      swap(x.blocks[i], x.blocks[j]);
      swap(x.factors[i], x.factors[j]);
    }
  }
  x.is_normalized = false;
  assert(x.check_inv());
}

F_Poly::F_Poly(dim_type d, Spec_Elem s, Topol t)
  : dim(d),
    empty(s == Spec_Elem::EMPTY),
    topol(t),
    blocks(),
    factors(),
    is_normalized(false) {
  if (is_empty())
    return;
  blocks.reserve(dim);
  for (dim_type i = 0; i != dim; ++i)
    blocks.emplace_back(1, i);
  factors.assign(dim, Factor(1, Spec_Elem::UNIVERSE, topology()));
  assert(check_inv());
}

bool
F_Poly::check_inv() const {
  std::string reason;
#ifdef NDEBUG
  auto maybe_dump = []() {};
#else // In debugging mode, be noisy.
  auto maybe_dump = [this, &reason]() {
    std::cerr << reason << std::endl;
    this->ascii_dump(std::cerr);
    std::cerr << reason << std::endl;
  };
#endif

  if (is_normalized) {
    auto copy = *this;
    copy.is_normalized = false;
    if (copy.hash() != hash()) {
      reason = "F_Poly broken: marked normalized when it is not.";
      maybe_dump();
      return false;
    }
  }

  if (dim < 0) {
    reason = "F_Poly broken: invalid space dimension";
    maybe_dump();
    return false;
  }

  // If it is empty, factors and blocks must be empty too.
  if (is_empty()) {
    if (!(factors.empty() && blocks.empty())) {
      reason = "F_Poly broken: empty polyhedron has factors";
      maybe_dump();
      return false;
    }
    // No other check for an empty polyhedron.
    return true;
  }

  // Blocks should not be empty.
  for (const auto& block : blocks) {
    if (block.empty()) {
      reason = "F_Poly broken: a block has space dim 0";
      maybe_dump();
      return false;
    }
  }

  // As many factors as blocks.
  if (factors.size() != blocks.size()) {
    reason = "F_Poly broken: #factors != #blocks";
    maybe_dump();
    return false;
  }

  // Blocks' sizes should match space dim.
  dim_type dim_ = 0;
  for (const auto& block : blocks)
    dim_ += block.size();
  if (dim != dim_) {
    reason = "F_Poly broken: space dimension mismatch";
    maybe_dump();
    return false;
  }

  // Each factor space dim should match its block.
  for (dim_type i = 0; i < num_rows(factors); ++i) {
    if (factors[i].space_dim() != num_rows(blocks[i])) {
      reason = "F_Poly broken: factor vs block space dim mismatch";
      maybe_dump();
      return false;
    }
  }

  // No factor is empty.
  if (std::any_of(factors.begin(), factors.end(),
                  std::mem_fn(&Poly::is_empty))) {
    reason = "F_Poly broken: empty factor";
    maybe_dump();
    return false;
  }

  // Each dimension appears once in blocks and there are no empty blocks.
  std::vector<bool> dims(dim, false);
  for (const auto& block : blocks) {
    if (block.size() == 0) {
      reason = "F_Poly broken: found empty block";
      maybe_dump();
      return false;
    }
    for (const auto d : block) {
      if (d < 0 || d >= dim) {
        reason = "F_Poly broken: block contains an illegal space dim";
        maybe_dump();
        return false;
      }
      if (dims[d]) {
        reason = "F_Poly broken: repeated space dim in blocks";
        maybe_dump();
        return false;
      }
      dims[d] = true;
    }
  }
  // Every dimension must appear in a block.
  if (std::find(dims.begin(), dims.end(), false) != dims.end()) {
    reason = "F_Poly broken: a space dim is missing from blocks";
    maybe_dump();
    return false;
  }

  // All tests passed.
  return true;
}

void
F_Poly::add_con(const Con& c) {
  if (is_empty())
    return;
  Block b = detail::extract_block(c);
  if (b.size() == 0) {
    if (c.is_tautological())
      return;
    else {
      assert(c.is_inconsistent());
      set_empty();
      return;
    }
  }
  auto i = merge(b);
  Con c_int = detail::convert(c, blocks[i]);
  factors[i].add_con(std::move(c_int));
  if (factors[i].is_empty())
    set_empty();
  assert(check_inv());
}

void
F_Poly::add_cons(const Cons& cs) { add_cons(cs.begin(), cs.end()); }

template <typename Iter>
void
F_Poly::add_cons(Iter first, Iter last, bool do_normalize) {
  if (first == last || is_empty()) {
    return;
  }

#if 1
  // PREVIOUS IMPLEMENTATION
  (void) do_normalize; // Avoid warnings
  for (auto it = first; it != last; ++it) {
    add_con(*it);
  }
  return;
#endif

  if (do_normalize)
    factorize();

  // EXPERIMENTAL:
  // Order constraints according to the interested blocks.
  using Iters = typename std::list<Iter>;
  std::map<dim_type, Iters> cs;

  for (auto it = first; it != last; ++it) {
    if (it->space_dim() == 0) {
      if (it->is_tautological())
        continue;
      else {
        assert(it->is_inconsistent());
        set_empty();
        return;
      }
    }
    auto num = detail::count_intersections(blocks, *it);
    cs[num].push_back(it);
  }

  // Add constraints ordered by number of intersections.
  for (const auto& cluster : cs) {
    for (auto it : cluster.second) {
      add_con(*it);
    }
  }
  assert(check_inv());
}

void
F_Poly::add_line_or_ray(const Gen& g) {
  assert(!is_empty() && g.is_line_or_ray());
  Block b_g = detail::extract_block(g);
  auto i = merge(b_g);
  auto g_int = detail::convert(g, blocks[i]);
  factors[i].add_gen(std::move(g_int));
  assert(check_inv());
}

Poly_Con_Rel
F_Poly::relation_with(const Con& c) const {
  if (is_empty())
    return Poly_Con_Rel::saturates()
      && Poly_Con_Rel::is_included()
      && Poly_Con_Rel::is_disjoint();

  if (c.is_inconsistent())
    return (c.is_strict_inequality() && c.inhomo_term() == 0)
      ? Poly_Con_Rel::saturates() && Poly_Con_Rel::is_disjoint()
      : Poly_Con_Rel::is_disjoint();

  if (c.is_tautological())
    return (c.is_equality() || c.inhomo_term() == 0)
      ? Poly_Con_Rel::saturates() && Poly_Con_Rel::is_included()
      : Poly_Con_Rel::is_included();

  assert(space_dim() > 0);
  Block b = detail::extract_block(c);
  assert(b.size() > 0);
  auto& x = const_cast<F_Poly&>(*this);
  auto i = x.merge(b);
  Con c_int = detail::convert(c, blocks[i]);
  return factors[i].relation_with(c_int);
}

Poly_Gen_Rel
F_Poly::relation_with(const Gen& g) const {
  assert(space_dim() >= g.space_dim());
  if (is_empty())
    return Poly_Gen_Rel::nothing();
  if (space_dim() == 0)
    return Poly_Gen_Rel::subsumes();

  if (g.is_line() || g.is_ray()) {
    Block b_g = detail::extract_block(g);
    for (auto i = 0, i_end = num_rows(blocks); i != i_end; ++i) {
      const auto& b = blocks[i];
      if (detail::are_disjoint(b, b_g))
        continue;
      Gen g_int = detail::convert(g, b);
      const auto& f = factors[i];
      if (f.relation_with(g_int) == Poly_Gen_Rel::nothing())
        return Poly_Gen_Rel::nothing();
    }
    return Poly_Gen_Rel::subsumes();
  }

  assert(g.is_point() || g.is_closure_point());
  for (auto i = 0, i_end = num_rows(blocks); i != i_end; ++i) {
    const auto& b = blocks[i];
    Gen g_int = detail::convert(g, b);
    const auto& f = factors[i];
    if (f.relation_with(g_int) == Poly_Gen_Rel::nothing())
      return Poly_Gen_Rel::nothing();
  }
  return Poly_Gen_Rel::subsumes();
}

bool
F_Poly::min_max(const bool minimizing,
                const Linear_Expr& expr, const Integer& inhomo,
                Integer& num, Integer& den, bool& included, Gen& g) const {
  if (space_dim() == 0) {
    if (is_empty())
      return false;
    num = inhomo;
    den = 1;
    included = true;
    g = point();
    return true;
  }
  if (is_empty())
    return false;

  Linear_Expr expr_g;
  Integer div_g = 1;
  Rational value = Rational(inhomo, 1);

  included = true;
  const Integer zero_i;
  Integer num_i, den_i, lcm, mult_g, mult_i;
  for (auto i = 0, i_end = num_rows(blocks); i != i_end; ++i) {
    const auto& b_i = blocks[i];
    bool included_i;
    Gen g_i = point();
    if (!factors[i].min_max(minimizing,
                            detail::convert(expr, b_i), zero_i,
                            num_i, den_i, included_i, g_i))
      return false;
    included = included && included_i;
    auto expr_i = detail::convert_back(g_i.linear_expr(), b_i);
    const auto& div_i = g_i.divisor();
    lcm_assign(lcm, div_g, div_i);
    exact_div_assign(mult_g, lcm, div_g);
    expr_g *= mult_g;
    exact_div_assign(mult_i, lcm, div_i);
    add_mul_assign(expr_g, mult_i, expr_i);
    div_g = std::move(lcm);
    value += Rational(num_i, den_i);
  }
  num = value.get_num();
  den = value.get_den();
  g = included ? point(expr_g, div_g) : closure_point(expr_g, div_g);
  return true;
}

bool
F_Poly::get_bounds(Var var,
                   Rational& min, Rational& max,
                   bool& min_inf, bool& max_inf) const {
  if (is_empty())
    return false;
  auto i = detail::get_block_index(blocks, var);
  Var v_int = detail::convert(var, blocks[i]);
  return factors[i].get_bounds(v_int, min, max, min_inf, max_inf);
}

bool
F_Poly::is_bounded_expr(bool from_below, const Linear_Expr& expr) const {
  if (space_dim() == 0 || is_empty())
    return true;
  for (auto i = 0, i_end = num_rows(blocks); i != i_end; ++i) {
    const auto& b_i = blocks[i];
    auto expr_i = detail::convert(expr, b_i);
    if (!factors[i].is_bounded_expr(from_below, expr_i))
      return false;
  }
  return true;
}

Cons
F_Poly::copy_cons() const {
  if (is_empty())
    return { Con::zero_dim_false() };
  if (space_dim() == 0)
    return Cons();

  Cons res;
  for (dim_type i = 0; i != num_rows(factors); ++i) {
    const auto& bi = blocks[i];
    for (const auto& c : factors[i].cons())
      res.push_back(detail::convert_back(c, bi));
  }
  return res;
}

Gens
F_Poly::copy_gens() const {
  if (is_empty())
    return Gens();

  if (space_dim() == 0)
    return { point() };

  assert(num_rows(factors) > 0);

  Block b = blocks[0];
  Factor f = factors[0];
  for (dim_type i = 1; i < num_rows(factors); ++i)
    detail::concat(b, f, blocks[i], factors[i]);

  f.map_space_dims(b);
  return f.copy_gens();
}

Cons
F_Poly::copy_normalized_cons() const {
  if (is_empty())
    return { Con::zero_dim_false() };
  if (space_dim() == 0)
    return Cons();

  Cons sing_rows;
  Cons sk_rows;
  for (dim_type i = 0; i != num_rows(factors); ++i) {
    const auto& bi = blocks[i];
    const auto& fi = factors[i];
    // Note: already materialized ns_rows in the local factor.
    for (const auto& c : fi.cons()) {
      if (c.is_equality())
        sing_rows.push_back(detail::convert_back(c, bi));
      else
        sk_rows.push_back(detail::convert_back(c, bi));
    }
  }
  // Expansion of detail::noralize for hash.
  detail::gauss(sing_rows);
  detail::back_substitute(sing_rows, sk_rows);

  sing_rows.insert(sing_rows.end(),
                   std::make_move_iterator(sk_rows.begin()),
                   std::make_move_iterator(sk_rows.end()));
  return sing_rows;
}

dim_type
F_Poly::num_min_cons() const {
  if (is_empty())
    return 1;
  dim_type res = 0;
  for (const auto& f : factors)
    res += f.num_min_cons();
  return res;
}

F_Poly::Gens_Info
F_Poly::gens_info() const {
  if (is_empty())
    return { 0, 0, 0, 0, 0 };

  dim_type ln = 0, r = 0, cp = 0, skp = 1, ns = 0;

  dim_type f_ln, f_r, f_cp, f_skp, f_ns;
  for (const auto& f : factors) {
    std::tie(f_ln, f_r, f_cp, f_skp, f_ns) = f.gens_info();
    ln += f_ln;
    r += f_r;
    cp = (cp * f_cp) + (skp * f_cp) + (cp * f_skp);
    skp *= f_skp;
    ns = (ns * f_ns) + (ns * f_skp) + (skp * f_ns);
  }
  return { ln, r, cp, skp, ns };
}

dim_type
F_Poly::num_min_gens() const {
  if (is_empty())
    return 0;
  if (space_dim() == 0)
    return 1;
  dim_type ln, r, cp, skp, ns;
  std::tie(ln, r, cp, skp, ns) = gens_info();
  return ln + r + cp + skp + ns;
}

bool
F_Poly::equals(const F_Poly& y) const {
  const auto& x = *this;
  if (x.space_dim() != y.space_dim())
    return false;
  if (x.is_empty())
    return y.is_empty();
  if (y.is_empty())
    return false;
  if (x.space_dim() == 0)
    return true;

  // check for luck
  if (x.num_min_cons() != y.num_min_cons())
    return false;
  if (x.gens_info() != y.gens_info())
    return false;

  Blocks lub = detail::least_upper_bound(x.blocks, y.blocks);
  detail::Refactor_Proxy x_fs(x.blocks, x.factors, lub);
  detail::Refactor_Proxy y_fs(y.blocks, y.factors, lub);
  for (auto i = num_rows(lub); i-- > 0; ) {
    if (!x_fs[i].equals(y_fs[i]))
      return false;
  }
  return true;
}

dim_type
F_Poly::affine_dim() const {
  dim_type res = 0;
  for (const auto& f : factors)
    res += f.affine_dim();
  return res;
}

void
F_Poly::set_empty() {
  blocks.clear();
  factors.clear();
  empty = true;
  assert(check_inv());
}

void
F_Poly::set_universe() {
  auto univ = F_Poly(space_dim(), topology());
  m_swap(univ);
}

namespace detail {

template <typename Hull_Fun>
dim_type
hull_factors(Blocks& bs1, Factors& fs1,
             const Refactor_Proxy& fs2, Hull_Fun hull_fun) {
  dim_type join_idx = not_a_dim();
  Index_Set joined_blocks;

  Factor fs2_join;
  const auto nb = num_rows(bs1);
  assert(nb > 0 && num_rows(fs1) == nb);
  for (auto i = 0; i != nb; ++i) {
    if (fs1[i] == fs2[i])
      continue;
    if (join_idx == not_a_dim()) {
      join_idx = i;
      fs2_join = fs2[i];
    }
    else {
      concat(bs1[join_idx], bs1[i]);
      fs1[join_idx].concatenate_assign(fs1[i]);
      fs2_join.concatenate_assign(fs2[i]);
      joined_blocks.set(i);
    }
  }

  if (join_idx != not_a_dim()) {
    erase_using_sorted_indices(bs1, joined_blocks);
    erase_using_sorted_indices(fs1, joined_blocks);
    // Safe: join_idx < all indices in joined_blocks
    hull_fun(fs1[join_idx], fs2_join);
  }
  return join_idx;
}

} // namespace detail

template <typename Hull_Fun>
void
F_Poly::hull_assign(const F_Poly& y, Hull_Fun hull_fun) {
  auto& x = *this;
  assert(x.space_dim() == y.space_dim());
  if (y.is_empty())
    return;
  if (x.is_empty() || x.space_dim() == 0) {
    x = y;
    return;
  }

  Blocks lub = detail::least_upper_bound(x.blocks, y.blocks);
  x.sync(lub);
  detail::Refactor_Proxy y_fs(y.blocks, y.factors, lub);

  auto join_idx = detail::hull_factors(x.blocks, x.factors, y_fs, hull_fun);
  if (join_idx != not_a_dim()) {
    // check for refactoring (only for join_idx)
    auto res = detail::factorize(x.blocks[join_idx], x.factors[join_idx]);
    detail::splice(x.blocks, res.first);
    detail::splice(x.factors, res.second);
  }
  assert(x.check_inv());
}

void
F_Poly::poly_hull_assign(const F_Poly& y) {
  hull_assign(y, std::mem_fn(&Factor::poly_hull_assign));
}

#if 0 // OLD CONSTRAINT HULL IMPLEMENTATION,
// May yield a different (often incomparable)
// result than the same operator on the corresponding Polys,
// due to the missing normalization of the factorization.
void
F_Poly::con_hull_assign(const F_Poly& y) {
  auto& x = *this;
  assert(x.space_dim() == y.space_dim());
  if (y.is_empty())
    return;
  if (x.is_empty() || x.space_dim() == 0) {
    x = y;
    return;
  }

  Blocks lub = detail::least_upper_bound(x.blocks, y.blocks);
  x.sync(lub);
  detail::Refactor_Proxy y_fs(y.blocks, y.factors, lub);
  for (auto i = num_rows(lub); i-- > 0; ) {
    auto& x_bi = x.blocks[i];
    auto& x_fi = x.factors[i];
    auto dim_i = x_fi.space_dim();
    bool normalizing_not_needed
      = std::is_sorted(x_bi.begin(), x_bi.end())
      || (x_fi.affine_dim() == dim_i)
      || (y_fs[i].affine_dim() == dim_i);
    if (normalizing_not_needed) {
      x_fi.con_hull_assign(y_fs[i]);
      continue;
    }
    // Sort block and compute permutation.
    std::sort(x_bi.begin(), x_bi.end());
    Dims perm(dim_i, not_a_dim());
    for (auto j = 0; j != dim_i; ++j) {
      auto iter = std::find(x_bi.begin(), x_bi.end(), lub[i][j]);
      assert(iter != x_bi.end());
      perm[j] = (iter - x_bi.begin());
    }
    // Permute space dims.
    x_fi.map_space_dims(perm);
    // Ditto (copy meant).
    auto y_fi = y_fs[i];
    y_fi.map_space_dims(perm);
    x_fi.con_hull_assign(y_fi);
  }
  assert(x.check_inv());
}
#endif // OLD CON_HULL IMPLEMENTATION

bool
F_Poly::constrains(Var v) const {
  if (is_empty())
    return true;
  auto i = detail::get_block_index(blocks, v);
  Var v_int = detail::convert(v, blocks[i]);
  return factors[i].constrains(v_int);
}

bool
F_Poly::contains(const F_Poly& y) const {
  const auto& x = *this;
  assert(x.check_inv() && y.check_inv());

  if (x.is_empty())
    return y.is_empty();
  if (y.is_empty())
    return true;
  if (x.space_dim() == 0)
    return true;

  Blocks lub = detail::least_upper_bound(x.blocks, y.blocks);
  detail::Refactor_Proxy x_fs(x.blocks, x.factors, lub);
  detail::Refactor_Proxy y_fs(y.blocks, y.factors, lub);
  for (auto i = num_rows(lub); i-- > 0; ) {
    if (!x_fs[i].contains(y_fs[i]))
      return false;
  }
  return true;
}

bool
F_Poly::strictly_contains(const F_Poly& y) const {
  const auto& x = *this;
  assert(x.check_inv() && y.check_inv());

  if (x.is_empty())
    return false;
  if (y.is_empty())
    return true;
  if (x.space_dim() == 0)
    return false;

  Blocks lub = detail::least_upper_bound(x.blocks, y.blocks);
  detail::Refactor_Proxy x_fs(x.blocks, x.factors, lub);
  detail::Refactor_Proxy y_fs(y.blocks, y.factors, lub);
  bool strictly = false;
  for (auto i = num_rows(lub); i-- > 0; ) {
    if (!x_fs[i].contains(y_fs[i]))
      return false;
    if (!strictly)
      strictly = !y_fs[i].contains(x_fs[i]);
  }
  return strictly;
}

bool
F_Poly::is_disjoint_from(const F_Poly& y) const {
  F_Poly z = *this;
  assert(z.check_inv() && y.check_inv());
  z.intersection_assign(y);
  return z.is_empty();
}

void
F_Poly::minimize() const {
  auto& x = const_cast<F_Poly&>(*this);
  for (auto& f : x.factors) {
    f.minimize();
    if (f.is_empty()) {
      x.set_empty();
      return;
    }
  }
}

namespace detail {

void ascii_dump_block(const Block& b, std::ostream& s) {
  const auto dim = num_rows(b);
  s << dim << " : {";
  for (dim_type j = 0; j != dim; ++j)
    s << " " << b[j];
  s << " }";
}

bool ascii_load_block(Block& b, std::istream& s) {
  dim_type dim = not_a_dim();
  if (not ((s >> dim) && (dim >= 0)))
    return false;
  b.resize(dim);
  std::string str;
  if (not (ascii_load_string(s, ":") && ascii_load_string(s, "{")))
    return false;
  for (dim_type j = 0; j != dim; ++j) {
    if (not ((s >> b[j]) && (b[j] >= 0)))
      return false;
  }
  if (not ascii_load_string(s, "}"))
    return false;
  return true;
}

} // namespace detail

void
F_Poly::ascii_dump(std::ostream& s) const {
  using namespace IO_Operators;
  s << "dim " << dim << '\n';
  s << "empty " << empty << '\n';
  s << "topol " << topol << '\n';

  s << "blocks " << num_rows(blocks) << std::endl;
  for (dim_type i = 0; i != num_rows(blocks); ++i) {
    detail::ascii_dump_block(blocks[i], s);
    s << '\n';
  }

  s << "factors " << num_rows(factors) << std::endl;
  for (dim_type i = 0; i < num_rows(factors); ++i) {
    s << "factor " << i << " ( ";
    if (i >= num_rows(blocks))
      s << "missing block -- broken F_Poly ";
    else {
      s << "block = ";
      detail::ascii_dump_block(blocks[i], s);
    }
    s << " )\n";
    factors[i].ascii_dump(s);
    s << '\n';
  }
}

bool
F_Poly::ascii_load(std::istream& s) {
  std::string str;
  if (not (ascii_load_string(s, "dim") && (s >> dim) && (dim >= 0)))
    return false;
  if (not (ascii_load_string(s, "empty") && (s >> empty)))
    return false;
  if (not (ascii_load_string(s, "topol") && (s >> str)))
    return false;
  if (str == "CLOSED")
    topol = Topol::CLOSED;
  else if (str == "NNC")
    topol = Topol::NNC;
  else
    return false;

  dim_type num_blocks = not_a_dim();
  if (not (ascii_load_string(s, "blocks") && (s >> num_blocks)))
    return false;
  if (num_blocks < 0)
    return false;
  blocks.resize(num_blocks);
  for (auto i = 0; i != num_blocks; ++i) {
    if (not detail::ascii_load_block(blocks[i], s))
      return false;
  }

  dim_type num_factors = not_a_dim();
  if (not (ascii_load_string(s, "factors") && (s >> num_factors)))
    return false;
  if (num_factors < 0)
    return false;
  factors.resize(num_factors);
  for (auto i = 0; i != num_factors; ++i) {
    dim_type index = not_a_dim();
    if (not (ascii_load_string(s, "factor") && (s >> index)))
      return false;
    if (not (index == i))
      return false;
    Block dummy;
    if (not (ascii_load_string(s, "(")
             && ascii_load_string(s, "block")
             && ascii_load_string(s, "=")
             && detail::ascii_load_block(dummy, s)
             && ascii_load_string(s, ")")))
      return false;
    if (dummy != blocks[i])
      return false;
    if (not factors[i].ascii_load(s))
      return false;
  }

  assert(check_inv());
  return true;
}

void
F_Poly::add_space_dims(dim_type m, bool project) {
  assert(m >= 0);

  if (m == 0)
    return;
  if (is_empty()) {
    dim += m;
    return;
  }

  Factor f(1, Spec_Elem::UNIVERSE, topology());
  if (project)
    f.add_con(Var(0) == 0);
  factors.insert(factors.end(), m, f);

  for (dim_type i = 0; i != m; ++i) {
    blocks.emplace_back(1, dim);
    ++dim;
  }
  assert(check_inv());
}

void
F_Poly::remove_higher_space_dims(dim_type new_dim) {
  if (is_empty()) {
    dim = new_dim;
    return;
  }
  for (dim_type i = space_dim(); i-- > new_dim; )
    remove_space_dim(Var(i));
}

void
F_Poly::map_space_dims(const Dims& pfunc) {
  assert(space_dim() == num_rows(pfunc));
  if (space_dim() == 0)
    return;

  Index_Set tbr;
  dim_type tbr_count = 0;
  for (auto i = 0, i_end = space_dim(); i != i_end; ++i) {
    if (pfunc[i] == not_a_dim()) {
      tbr.set(i);
      ++tbr_count;
    }
  }

  if (is_empty()) {
    dim -= tbr_count;
    return;
  }

  auto permute_dims = [](Blocks& bs, const Dims& pf) {
    for (auto& b : bs)
      for (auto& i : b)
        i = pf[i];
  };

  if (tbr_count == 0) {
    permute_dims(blocks, pfunc);
    assert(check_inv());
    return;
  }

  assert(tbr_count > 0);
  remove_space_dims(tbr);
  // Recompute partial function after removal.
  Dims perm = pfunc;
  erase_using_sorted_indices(perm, tbr);
  permute_dims(blocks, perm);
  assert(check_inv());
}

void
F_Poly::fold_space_dims(const Index_Set& vars, Var dest) {
  assert(!vars.test(dest.id()));

  if (is_empty()) {
    dim -= vars.size();
    return;
  }

  Block b(vars.begin(), vars.end());
  b.push_back(dest.id());

  auto i = merge(b);
  auto& b_i = blocks[i];
  auto& f_i = factors[i];

  Var c_dest = detail::convert(dest, b_i);

  Index_Set c_vars;
  for (auto v_j : vars) {
    Var c_var = detail::convert(Var(v_j), b_i);
    c_vars.set(c_var.id());
  }

  // Fold dimensions in factor f_i
  f_i.fold_space_dims(c_vars, c_dest);
  // Remove folded dimensions from b_i
  erase_using_sorted_indices(b_i, c_vars);

  // Adjust blocks: iterating backwards is meant.
  for (auto it = vars.end(), it_end = vars.begin(); it != it_end; ) {
    --it;
    detail::reduce_blocks(blocks, Var(*it));
  }

  // Update space dimension
  dim -= vars.size();

  assert(check_inv());
}

void
F_Poly::expand_space_dim(Var v, dim_type m) {
  if (m == 0)
    return;
  if (is_empty()) {
    dim += m;
    return;
  }

  auto idx = detail::get_block_index(blocks, v);
  const auto old_dim = dim;
  dim += m;
  if (blocks[idx].size() == 1) {
    // Add m copies of factor idx (and corresponding blocks).
    for (dim_type d = 0; d != m; ++d)
      blocks.emplace_back(1, old_dim + d);
    // Note: this reserve avoids UB.
    factors.reserve(factors.size() + m);
    factors.insert(factors.end(), m, factors[idx]);
  } else {
    Var v_int = detail::convert(v, blocks[idx]);
    for (dim_type d = 0; d != m; ++d)
      blocks[idx].push_back(old_dim + d);
    factors[idx].expand_space_dim(v_int, m);
  }
  assert(check_inv());
}

void
F_Poly::unconstrain(Var v) {
  if (is_empty())
    return;
  auto i = detail::get_block_index(blocks, v);
  auto& b_i = blocks[i];
  auto& f_i = factors[i];
  if (num_rows(b_i) == 1)
    f_i.unconstrain(Var(0));
  else {
    Var v_int = detail::convert(v, b_i);
    b_i.erase(b_i.begin() + v_int.id());
    f_i.remove_space_dim(v_int);
    // Add the universe 1-dim factor.
    blocks.emplace_back(1, v.id());
    factors.emplace_back(1, topology());
  }
  assert(check_inv());
}

void
F_Poly::remove_space_dim(Var v) {
  assert(v.space_dim() <= space_dim());
  if (is_empty()) {
    --dim;
    return;
  }
  auto i = detail::get_block_index(blocks, v);
  auto& b_i = blocks[i];
  if (num_rows(b_i) == 1) {
    blocks.erase(blocks.begin() + i);
    factors.erase(factors.begin() + i);
  } else {
    Var v_int = detail::convert(v, b_i);
    b_i.erase(b_i.begin() + v_int.id());
    factors[i].remove_space_dim(v_int);
    is_normalized = false;
  }
  detail::reduce_blocks(blocks, v);
  --dim;
  assert(check_inv());
}

void
F_Poly::intersection_assign(const F_Poly& y) {
  assert(space_dim() == y.space_dim());
  if (is_empty())
    return;
  if (y.is_empty())
    set_empty();
  else
    add_cons(y.copy_cons());
}

void
F_Poly::poly_difference_assign(const F_Poly& y) {
  auto& x = *this;
  if (x.is_empty() || y.is_empty())
    return;
  if (x.space_dim() == 0) {
    x.set_empty();
    return;
  }
  // Factorization is not really needed, just a speculative optimization.
  x.factorize();
  y.factorize();
  Blocks lub = detail::least_upper_bound(x.blocks, y.blocks);
  detail::Refactor_Proxy x_fs(x.blocks, x.factors, lub);
  detail::Refactor_Proxy y_fs(y.blocks, y.factors, lub);

  // Compute containment for each factor.
  Index_Set contained;
  for (auto i = num_rows(lub); i-- > 0; ) {
    if (y_fs[i].contains(x_fs[i]))
      contained.set(i);
  }
  if (contained.size() == num_rows(lub)) {
    x.set_empty();
    return;
  }
  // Compute factor-wise complement
  for (auto i = num_rows(lub); i-- > 0; ) {
    x.factors[i] = x_fs[i];
    if (!contained.test(i))
      x.factors[i].poly_difference_assign(y_fs[i]);
  }
  x.blocks = std::move(lub);
  assert(x.check_inv());
}

void
F_Poly::concatenate_assign(const F_Poly& y) {
  auto& x = *this;
  assert(x.topol == y.topol);
  if (x.is_empty() || y.is_empty()) {
    x.dim += y.dim;
    x.set_empty();
    return;
  }
  if (y.space_dim() == 0)
    return;
  if (x.space_dim() == 0) {
    x = y;
    return;
  }

  // Copy before modifying.
  const auto x_dim = x.dim;
  const auto x_nb = num_rows(x.blocks);

  x.dim += y.dim;
  x.blocks.insert(x.blocks.end(),
                  y.blocks.begin(), y.blocks.end());
  for (auto i = x_nb, i_end = num_rows(x.blocks); i != i_end; ++i)
    for (auto& d : x.blocks[i])
      d += x_dim;
  x.factors.insert(x.factors.end(),
                   y.factors.begin(), y.factors.end());
  assert(x.check_inv());
}

void
F_Poly::affine_image(Var var, const Linear_Expr& expr,
                     const Integer& inhomo, const Integer& den) {
  if (is_empty())
    return;

  Block b = detail::extract_block(expr);
  detail::add_var(b, var);
  auto i = merge(b);
  const auto& b_i = blocks[i];

  Var v_int = detail::convert(var, b_i);
  Linear_Expr le_int = detail::convert(expr, b_i);
  factors[i].affine_image(v_int, le_int, inhomo, den);
  assert(check_inv());
}

void
F_Poly::affine_preimage(Var var, const Linear_Expr& expr,
                        const Integer& inhomo, const Integer& den) {
  if (is_empty())
    return;

  Block b = detail::extract_block(expr);
  detail::add_var(b, var);
  auto i = merge(b);
  const auto& b_i = blocks[i];

  Var v_int = detail::convert(var, b_i);
  Linear_Expr le_int = detail::convert(expr, b_i);
  factors[i].affine_preimage(v_int, le_int, inhomo, den);
  assert(check_inv());
}

void
F_Poly::parallel_affine_image(const Vars& vars,
                              const Linear_Exprs& exprs,
                              const Integers& inhomos,
                              const Integers& dens) {
  detail::par_affine_image_aux(*this, vars, exprs, inhomos, dens);
}

Poly
F_Poly::get_poly() const {
  if (space_dim() == 0 || is_empty())
    return Poly(space_dim(),
                (is_empty()) ? Spec_Elem::EMPTY : Spec_Elem::UNIVERSE,
                topology());

  Block b(space_dim());
  std::iota(b.begin(), b.end(), 0);
  Index_Set joining;
  joining.set_until(num_rows(blocks));
  return detail::join_and_remap(blocks, factors, joining, b);
}

void
F_Poly::widening_assign(const F_Poly& y, widen_type widen,
                        const Cons* upto_ptr) {
  auto& x = *this;
  assert(x.topology() == y.topology());
  assert(x.space_dim() == y.space_dim());
  if (x.space_dim() == 0 || x.is_empty() || y.is_empty())
    return;

  Index_Set upto;
  if (upto_ptr) {
    // Before modifying x, select from upto_cs those constraints
    // that are satisfied by x.
    const auto& upto_cs = *upto_ptr;
    for (dim_type i = upto_cs.size(); i-- > 0; ) {
      auto rel = x.relation_with(upto_cs[i]);
      if (rel.implies(Poly_Con_Rel::is_included()))
        upto.set(i);
    }
  }

  // FIXME: temporary, could be improved?
  Blocks lub = detail::least_upper_bound(x.blocks, y.blocks);
  x.sync(lub);
  detail::Refactor_Proxy y_fs(y.blocks, y.factors, lub);
  for (auto i = num_rows(lub); i-- > 0; )
    x.factors[i].widening_assign(y_fs[i], widen);

  if (upto_ptr) {
    // Now add selected upto constraints.
    const auto& upto_cs = *upto_ptr;
    for (auto i : upto)
      x.add_con(upto_cs[i]);
  }
  assert(x.check_inv());
}

BBox
F_Poly::get_bounding_box() const {
  if (is_empty())
    return BBox(space_dim(), Spec_Elem::EMPTY);
  BBox res(space_dim());
  for (auto i = num_rows(blocks); i-- > 0; ) {
    BBox bbox = factors[i].get_bounding_box();
    const auto& block = blocks[i];
    // Move factor bounds into shell bounds.
    for (auto fd = num_rows(block); fd-- > 0; ) {
      auto sd = block[fd];
      res.bounds[sd] = std::move(bbox.bounds[fd]);
    }
  }
  res.recompute_pseudo_volume();
  return res;
}

F_Poly
F_Poly::split(const Con& c, Topol t) {
  // Init result with empty poly
  auto res = F_Poly(space_dim(), Spec_Elem::EMPTY, topology());

  if (is_empty())
    return res;
  Block b = detail::extract_block(c);
  if (b.size() == 0) {
    if (c.is_tautological())
      return res;
    assert(c.is_inconsistent());
    m_swap(res);
    return res;
  }

  auto i = merge(b);
  auto& bi = blocks[i];
  auto& fi = factors[i];
  assert(!fi.is_empty());
  Con c_int = detail::convert(c, bi);
  auto res_fi = fi.split(c_int, t);
  if (res_fi.is_empty()) {
    assert(!fi.is_empty());
    return res;
  }
  if (fi.is_empty()) {
    m_swap(res);
    res.factors[i] = std::move(res_fi);
    return res;
  }
  // Both fi and res_fi not empty
  res = *this;
  res.factors[i] = std::move(res_fi);
  return res;
}

void
F_Poly::time_elapse_assign(const F_Poly& y) {
  auto& x = *this;
  assert(x.space_dim() == y.space_dim());
  if (x.is_empty())
    return;
  if (y.is_empty()) {
    x.set_empty();
    return;
  }
  if (x.space_dim() == 0)
    return;

  Gens y_rays;
  detail::add_as_rays(y.copy_gens(), y_rays);
  Index_Set y_vars;
  for (const auto& g : y_rays) {
    const auto& ex = g.linear_expr();
    for (auto i = ex.space_dim(); i-- > 0; )
      if (!ex.get(i).is_zero())
        y_vars.set(i);
  }
  // Check if y was (just) the origin of the vecor space.
  if (y_vars.empty())
    return;

  Block b(y_vars.begin(), y_vars.end());
  auto i = merge(b);
  auto& x_bi = x.blocks[i];
  auto& x_fi = x.factors[i];
  for (const auto& g : y_rays)
    x_fi.add_gen(detail::convert(g, x_bi));
  auto res = detail::factorize(x_bi, x_fi);
  detail::splice(x.blocks, res.first);
  detail::splice(x.factors, res.second);
  assert(x.check_inv());
}

void
F_Poly::normalize() const {
  if (is_normalized)
    return;

  // Only semantically const
  auto& x = const_cast<F_Poly&>(*this);

  // Compute maximal factorization
  x.factorize();

  // helper
  auto blocks_lex_lesser =
    [](const Block& x, const Block& y) {
      for (auto i = 0; i < num_rows(x); ++i) {
        if (i > num_rows(y) || x[i] > y[i])
          return false;
        if (x[i] < y[i])
          return true;
      }
      return false;
    };

  // Sort blocks
  detail::synchronous_sort(x.blocks, x.factors, blocks_lex_lesser);

  x.is_normalized = true;
}

size_t
F_Poly::hash() const {
  normalize();
  size_t res = hash_size(space_dim());
  for (const auto& b : blocks) {
    for (auto i : b)
      hash_combine(res, i);
  }
  for (const auto& f : factors)
    hash_combine(res, f.hash());
  return res;
}

void
F_Poly::print(std::ostream& os) const {
#if 1
  Poly ph = get_poly();
  ph.minimize();
  ph.print(os);
#else
  if (is_empty())
    os << "false";
  minimize();
  bool comma = false;
  for (const auto& c : cons()) {
    if (comma) os << ", ";
    using namespace IO_Operators;
    os << c;
    comma = true;
  }
#endif
}

} // namespace pplite
