/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pplite-config.h"
#include "Linear_Expr.hh"
#include "Low_Level_Stats.hh"
#include "ascii_dump_load.hh"

#include <iostream>
#include <algorithm>

namespace pplite {

int
compare(Linear_Expr const& x, Linear_Expr const& y) {
  const dim_type min_dim = std::min(x.space_dim(), y.space_dim());
  for (dim_type i = 0; i < min_dim; ++i) {
    auto res = compare(x.get(i), y.get(i));
    if (res != 0)
      return res;
  }
  for (dim_type i = min_dim, i_end = x.space_dim(); i != i_end; ++i) {
    auto res = sgn(x.get(i));
    if (res != 0)
      return res;
  }
  for (dim_type i = min_dim, i_end = y.space_dim(); i != i_end; ++i) {
    auto res = -sgn(y.get(i));
    if (res != 0)
      return res;
  }
  return 0;
}

Integer
Linear_Expr::gcd(dim_type first, dim_type last) const {
  assert(0 <= first && first <= last && last <= space_dim());
  auto i = first_nonzero(first, last);
  if (i == last)
    return Integer::zero();
  const auto& x = impl();
  Integer res = x[i];
  if (sgn(res) < 0)
    neg_assign(res);
  if (res == 1)
    return res;
  for (++i; i != last; ++i) {
    if (x[i] != 0) {
      gcd_assign(res, res, x[i]);
      if (res == 1)
        return res;
    }
  }
  return res;
}

void
Linear_Expr::normalize(Integer& inhomo) {
  Integer gcd_in = gcd(0, space_dim());
  if (inhomo != 0 && gcd_in != 1) {
    if (gcd_in != 0)
      gcd_assign(gcd_in, gcd_in, inhomo);
    else {
      gcd_in = inhomo;
      if (sgn(gcd_in) < 0)
        neg_assign(gcd_in);
    }
  }
  if (gcd_in == 0 || gcd_in == 1)
    return;

  // Divide the coefficients by the GCD.
  exact_div_assign(inhomo, inhomo, gcd_in);
  auto& x = impl();
  for (auto i = x.size(); i-- > 0; ) {
    auto& x_i = x[i];
    exact_div_assign(x_i, x_i, gcd_in);
  }
}

void
Linear_Expr::linear_combine(Linear_Expr& x, const Linear_Expr& y,
                            const Integer& x_coeff, const Integer& y_coeff,
                            Integer& x_inhomo, const Integer& y_inhomo) {
#if PPLITE_LOW_LEVEL_COUNTERS
  LLOp_Stats::bump_linear_comb();
#endif
  assert(x_coeff != 0 && y_coeff != 0);
  // Handle inhomo terms.
  x_inhomo *= x_coeff;
  add_mul_assign(x_inhomo, y_inhomo, y_coeff);
  // Handle homo terms.
  // FIXME: optimize for special cases (x/y_coeff == +/- 1).
  auto min_dim = std::min(x.space_dim(), y.space_dim());
  for (auto i = 0; i != min_dim; ++i) {
    x[i] *= x_coeff;
    if (y[i] == 0)
      continue;
    add_mul_assign(x[i], y[i], y_coeff);
  }
  if (x.space_dim() > min_dim) {
    auto max_dim = x.space_dim();
    for (auto i = min_dim; i != max_dim; ++i)
      x[i] *= x_coeff;
  } else if (y.space_dim() > min_dim) {
    auto max_dim = y.space_dim();
    x.set_space_dim(max_dim);
    for (auto i = min_dim; i != max_dim; ++i) {
      if (y[i] == 0)
        continue;
      x[i] = y[i];
      x[i] *= y_coeff;
    }
  }
}

void Linear_Expr::print(std::ostream& os) const {
  if (is_zero()) {
    os << "0";
    return;
  }

  bool first = true;
  auto sd = space_dim();
  for (int i = 0; i < sd; ++i) {
    auto s = sgn(row[i]);
    if (s == 0)
      continue;
    if (s > 0) {
      if (!first)
        os << " + ";
      if (row[i] != 1) {
        row[i].print(os);
        os << "*";
      }
    } else {
      assert(s < 0);
      if (first)
        os << "-";
      else
        os << " - ";
      if (row[i] != -1) {
        neg(row[i]).print(os);
        os << "*";
      }
    }
    first = false;
    Var(i).print(os);
  }
}

void
Linear_Expr::ascii_dump(std::ostream& s) const {
  using IO_Operators::operator<<;
  s << "dim " << space_dim() << " : ";
  for (dim_type i = 0; i < space_dim(); ++i) {
    get(i).print(s);
    s << ' ';
  }
}

bool
Linear_Expr::ascii_load(std::istream& s) {
  dim_type dim = 0;
  if (!(ascii_load_string(s, "dim")
        && (s >> dim)
        && (dim >= 0)
        && ascii_load_string(s, ":")))
    return false;
  row.resize(dim);
  return std::all_of(row.begin(), row.end(),
                     [&s](Integer& i) { return i.ascii_load(s); });
}

} // namespace pplite
