/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pplite-config.h"
#include "Sat.hh"

#include <string>

namespace pplite {

void
Sat::resize(dim_type new_nrows, dim_type new_ncols) {
  Impl& x = impl();
  x.rows.resize(new_nrows);
  if (new_ncols < x.row_size) {
    for (auto& r : x.rows)
      r.clear_from(new_ncols);
  }
  x.row_size = new_ncols;
  assert(check_inv());
}

void
Sat::transpose_assign(const Sat& y) {
  const auto y_nrows = y.num_rows();
  const auto y_ncols = y.num_cols();
  Sat tmp(y_ncols, y_nrows);
  for (auto i = y_nrows; i-- > 0; ) {
    for (auto j = y[i].last(); j != Row::end_pos; j = y[i].prev(j)) {
      tmp[j].set(i);
    }
  }
  m_swap(tmp);
  assert(check_inv());
}

void
Sat::ascii_dump(std::ostream& s) const {
  const char sep = ' ';
  s << num_rows() << sep << 'x' << sep << num_cols() << "\n";
  for (const auto& row : rows()) {
    for (dim_type j = 0; j < num_cols(); ++j) {
      s << row[j] << sep;
    }
    s << "\n";
  }
}

bool
Sat::ascii_load(std::istream& s) {
  auto& x = *this;
  dim_type nrows;
  dim_type ncols;
  std::string str;
  if (!(s >> nrows))
    return false;
  if (!(s >> str) || str != "x")
    return false;
  if (!(s >> ncols))
    return false;
  resize(nrows, ncols);

  for (dim_type i = 0; i < num_rows(); ++i) {
    for (dim_type j = 0; j < num_cols(); ++j) {
      int bit;
      if (!(s >> bit))
        return false;
      if (bit != 0)
        x[i].set(j);
      else
        x[i].clear(j);
    }
  }
  assert(check_inv());
  return true;
}

bool
Sat::check_inv() const {
#ifndef NDEBUG
  using std::endl;
  using std::cerr;
#endif

  const auto& x = *this;
  for (auto i = num_rows(); i-- > 0; ) {
    const auto& row = x[i];
    if (!row.check_inv())
      return false;
    else if (row.last() != Row::end_pos && row.last() >= num_cols()) {
#ifndef NDEBUG
      cerr << "Sat[" << i << "] is a row with too many bits!"
           << endl
           << "(row_size == " << impl().row_size
           << ", row.last() == " << row.last() << ")"
           << endl;
#endif
      return false;
    }
  }
  return true;
}

} // namespace pplite

