/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pplite-config.h"
#include "clock.hh"

#include <cerrno>
#include <cstdlib>
#include <cstring>
#include <iostream>

namespace pplite {

void Clock::restart() {
  using namespace std;
#if USE_RUSAGE
  struct rusage usage;
  if (getrusage(RUSAGE_SELF, &usage) != 0) {
    cerr << "getrusage failed: " << strerror(errno) << endl;
    exit(1);
  }
  saved_time = usage.ru_utime;
#else
  timespec ctime;
  if (clock_gettime(CLOCK_MONOTONIC, &ctime) != 0) {
    cerr << "clock_gettime failed: " << strerror(errno) << endl;
    exit(1);
  }
  saved_time = ctime;
#endif
}

std::pair<time_t, time_t>
Clock::elapsed_time() const {
  using namespace std;
#if USE_RUSAGE
  struct rusage usage;
  if (getrusage(RUSAGE_SELF, &usage) != 0) {
    cerr << "getrusage failed: " << strerror(errno) << endl;
    exit(1);
  }
  const time_t current_secs = usage.ru_utime.tv_sec;
  const time_t current_usecs = usage.ru_utime.tv_usec;
  const time_t saved_secs = saved_time.tv_sec;
  const time_t saved_usecs = saved_time.tv_usec;
#else
  struct timespec ctime;
  if (clock_gettime(CLOCK_MONOTONIC, &ctime) != 0) {
    cerr << "clock_gettime failed: " << strerror(errno) << endl;
    exit(1);
  }
  const time_t current_secs = ctime.tv_sec;
  const time_t current_usecs = ctime.tv_nsec / 1000;
  const time_t saved_secs = saved_time.tv_sec;
  const time_t saved_usecs = saved_time.tv_nsec / 1000;
#endif
  bool wrapped = (current_usecs < saved_usecs);
  time_t secs = current_secs - saved_secs - (wrapped ? 1 : 0);
  time_t usecs = (wrapped ? 1000000 : 0) + current_usecs - saved_usecs;
  return std::make_pair(secs, usecs);
}

} // namespace pplite
