/* PPLite: a lightweight library for convex polyhedra derived from PPL.
   Copyright (C) 2018-2020 Enea Zaffanella <enea.zaffanella@unipr.it>

This file is part of PPLite.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "pplite-config.h"

#include "globals.hh"
#include "utils.hh"
#include "support_utils.hh"
#include "Bit_Row.hh"
#include "Con.hh"
#include "Gen.hh"
#include "Index_Partition.hh"
#include "Index_Set.hh"
#include "Integer.hh"
#include "Integer.hh"
#include "Linear_Expr.hh"
#include "Poly.hh"
#include "Poly_templ.hh"
#include "Poly_min.hh"
#include "Sat.hh"
#include "Scalar_Prod.hh"

#include <algorithm>
#include <cassert>
#include <utility>
#include <vector>

namespace pplite {
namespace detail {

using Empty_Pair_Res = std::pair<bool, bool>;

bool
has_no_eq_point(const Ranges& sk_ranges,
                const Gens& sk_strict,
                const NS_Rows& ns_rows) {
  Index_Set sk_neg, sk_eq, sk_pos;
  sk_ranges_to_index_sets(sk_ranges, sk_neg, sk_eq, sk_pos);
  for (const auto& g : sk_eq)
    if (sk_strict[g].is_point())
      return false;
  return !std::any_of(ns_rows.cbegin(), ns_rows.cend(),
                      [&sk_eq](const Index_Set& ns) {
                        return subset_or_equal(ns, sk_eq);
                      });

}

/*******************************************************/
// Process skeleton

// Combines the skeleton generator in SK+ with those in SK-.
// Creates new elements in sk_strict and sk_nonstrict.
// If nnc_split, performs the COPY of sk,
// Otherwise, updates only the `strict' side.
// Updates only the `sat_strict': the copy is posponed.
void
split_combine_sk_rows(Ranges& sk_ranges, Gens& sk_strict, Gens& sk_nonstrict,
                      Integers& sp, Sat& sat_strict,
                      const bool nnc_split,
                      const dim_type min_sat, const dim_type max_sat) {
  if (nnc_split) {
    // COPY sk_rows
    // The original sk is shared,
    // only the new elements to be added have different types.
    sk_nonstrict = sk_strict;
  }

  Integer norm_sp_pos;
  Integer norm_sp_neg;
  const dim_type old_num_sk_rows = sk_strict.size();
  Sparse_Index_Partition<Range::POS> sk_pos(sk_ranges);
  Sparse_Index_Partition<Range::NEG> sk_neg(sk_ranges);
  Dims sat_ones(sk_strict.size(), 0);
  for (auto i = sat_ones.size(); i-- > 0; ) {
    if (sk_ranges[i] != Range::EQ)
      sat_ones[i] = sat_strict[i].count_ones();
  }

  for (auto pos : sk_pos) {
    for (auto neg : sk_neg) {
      // Exploit saturation information to perform adjacency tests.
      // Note: delay creation of `new_row' and `new_satrow'.
      const dim_type new_satrow_ones
        = sat_strict[pos].count_ones_in_union(sat_strict[neg]);
      if (quick_non_adj_test(min_sat, max_sat, new_satrow_ones))
        continue;
      Bit_Row new_satrow(sat_strict[pos], sat_strict[neg]);
      if (not quick_adj_test(pos, neg, new_satrow_ones, sat_ones) &&
          not combinatorial_adj_test(0, old_num_sk_rows,
                                     pos, neg, new_satrow, sat_strict))
        continue;

      // Here `pos' and `neg' are adjacent: update `sat_strict'.
      sat_strict.add_row(std::move(new_satrow));

      // Add the new row to `sk_strict'
      auto new_row_strict = sk_strict[neg];
      // Combination can be done once.
      combine_row(new_row_strict, sk_strict[pos],
                  sp[neg], sp[pos],
                  norm_sp_neg, norm_sp_pos);
      new_row_strict.set_type(combine_type(sk_strict[pos].type(),
                                           sk_strict[neg].type(),
                                           nnc_split));
      if (nnc_split) {
        // Copy the new element
        auto new_row_nonstrict = new_row_strict;
        // Set the right type, according to the side.
        new_row_nonstrict.set_type(combine_type(sk_strict[neg].type(),
                                                sk_strict[pos].type(),
                                                false));
        sk_nonstrict.push_back(std::move(new_row_nonstrict));
      }
      sk_strict.push_back(std::move(new_row_strict));

      assert(num_rows(sk_strict) == sat_strict.num_rows());
    } // end loop on sk_neg
  } // end loop on sk_pos

  sp.resize(sk_strict.size());
  sk_ranges.resize(sk_strict.size(), Range::EQ);
}


/**
 * Computes the split sk systems, copying of the original one.
 * NB: sk_[non]strict is consistent only if !empty_[non]strict.
 * In that case, will be the caller to set the empty DD.
 * Updates only the `sat_strict': the copy is postponed.
 */
Empty_Pair_Res
split_skel(dim_type space_dim, dim_type num_lines, dim_type num_eq,
           Gens& sk_strict, Gens& sk_nonstrict,
           Ranges& sk_ranges, const Con& beta_strict,
           const NS_Rows& ns_strict,
           Sat& sat_strict) {
  assert(beta_strict.is_inequality());
  const bool nnc_split = beta_strict.is_strict_inequality();
  bool empty_strict = false;
  bool empty_nonstrict = false;

  Integers scal_prods;
  compute_scalar_prods(beta_strict, sk_strict, scal_prods);

  // Create partition for skeleton rows.
  // The partition is relative to the strict side.
  sk_partition(scal_prods, sk_ranges);
  Index_Partition<Range::NEG> sk_neg(sk_ranges);
  Index_Partition<Range::EQ> sk_eq(sk_ranges);
  Index_Partition<Range::POS> sk_pos(sk_ranges);

  if (!sk_neg.empty() && !sk_pos.empty()) {
    // Here SK+ and SK- are non-empty.
    // Combine SK+ and SK- *adjacent* generators.
    const dim_type min_sat = space_dim - num_lines - 1;
    const dim_type max_sat = sat_strict.num_cols() + num_eq;
    split_combine_sk_rows(sk_ranges, sk_strict, sk_nonstrict,
                          scal_prods, sat_strict,
                          nnc_split, min_sat, max_sat);
  }
  else {
    const bool no_eq_point = has_no_eq_point(sk_ranges, sk_strict, ns_strict);
    if (sk_neg.empty()) {
      if (sk_eq.empty() || no_eq_point)
        empty_nonstrict = true;
      if (nnc_split && sk_pos.empty()) {
        // strict side is only in eq, but it is strict.
        empty_strict = true;
      }
      // The non-skel handling is postponed.
    }
    else {
      assert(sk_pos.empty());
      if (nnc_split || sk_eq.empty() || no_eq_point)
        empty_strict = true;
    }

    // Copy original gs system only if both sides are non-empty.
    // If !nnc_split the copy is posponed, since there could be
    // shared promotions of the same ns elements.
    if (!empty_strict && !empty_nonstrict && nnc_split)
      sk_nonstrict = sk_strict;
  }

  // Update sat_strict cols.
  if (!empty_strict && !empty_nonstrict) {
    sat_strict.add_cols(1);
    update_last_sat_column(sat_strict, sk_pos);
    update_last_sat_column(sat_strict, sk_neg);
  }

  assert(!empty_strict || !empty_nonstrict);
  return std::make_pair(empty_strict, empty_nonstrict);
}

/*******************************************************/
// Process non-skeleton

/* Move-ns method for NNC split */
// Computes the closures once and differentiates only the projection phase.
// Checks the supports to be promoted only in the nonstrict side.
void
split_move_ns(NS_Rows& ns_strict, NS_Rows& ns_nonstrict,
              Ranges& ns_ranges_strict, Ranges& ns_ranges_nonstrict,
              Gens& sk_nonstrict,
              const Index_Set& iset_neg,
              const Index_Set& iset_eq,
              const Index_Set& iset_pos,
              NS_Rows& nsstar_strict, NS_Rows& nsstar_nonstrict,
              const Sat& sat_c, Sat& sat_g) {
  // Originally ns_ranges are the same.
  Index_Partition<Range::POS_NEG> ns_pos_neg(ns_ranges_strict);
  Dims to_be_del;
  for (auto i : ns_pos_neg) {
    assert(ns_ranges_strict[i] == Range::POS_NEG);
    Index_Set& supp_strict = ns_strict[i];
    // Checkme: closures can be done once, since sat info are shared.
    support_closure(supp_strict, sat_c, sat_g);

    Index_Set supp_nonstrict = supp_strict;
    // Differentiate only the projection phase.
    supp_projection(supp_strict, true, iset_neg, iset_eq);
    supp_projection(supp_nonstrict, false, iset_pos, iset_eq);
    // No need to check if it has to be promoted
    assert(supp_strict.size() > 1);
    add_to_minimal_set_of_ns(supp_strict, nsstar_strict);

    // Check if it can be moved into SK.
    if (supp_nonstrict.size() == 1) {
      // Change its only supporter's type into point.
      const auto supporter_index = *(supp_nonstrict.begin());
      promote_singleton(sk_nonstrict[supporter_index]);
      to_be_del.push_back(i);
      // Consider next ns in pos_neg
      continue;
    }
    assert(supp_nonstrict.size() > 1);
    add_to_minimal_set_of_ns(supp_nonstrict, nsstar_nonstrict);
  }
  // Remove ns that has been moved into SK.
  erase_using_sorted_indices(ns_nonstrict, to_be_del);
  erase_using_sorted_indices(ns_ranges_nonstrict, to_be_del);
}

/** Create_ns method for NNC split. **/
// Add to NS*strict:
// - ns* from NEG projected in POS,
// - ns* from EQ projected in POS.
// Add to NS*nonstrict:
// - ns* from NEG projected in EQ,
// - ns* from POS projected in EQ.
void
split_create_ns(NS_Rows& ns_strict, NS_Rows& ns_nonstrict,
                Ranges& ns_ranges_strict, Ranges& ns_ranges_nonstrict,
                Gens& sk_strict, Gens& sk_nonstrict, const Ranges& sk_ranges,
                const Index_Set& iset_pos,
                const Index_Set& iset_eq,
                const Index_Set& iset_neg,
                NS_Rows& nsstar_strict, NS_Rows& nsstar_nonstrict,
                const Sat& sat_c, const Sat& sat_g,
                dim_type min_sat, dim_type max_sat,
                bool should_create_nonstrict) {
  // enumerate_faces from NEG can be done once:
  // Store the computed set, reproject it in EQ if needed.
  NS_Rows set_from_neg;
  create_ns_from_Q_with_face_enum<Range::NEG>(Range::POS,
                                              sk_strict, ns_strict,
                                              sk_ranges, ns_ranges_strict,
                                              set_from_neg,
                                              iset_pos, iset_eq, iset_neg,
                                              sat_c, sat_g,
                                              true);
  // Here `set_from_neg' stores the created ns projected in POS.
  for (auto& ns_star : set_from_neg) {
    // Add it to nsstar_strict (it is copied).
    add_to_minimal_set_of_ns(ns_star, nsstar_strict);
    if (should_create_nonstrict) {
      // Reproject it in EQ
      supp_projection(ns_star, false, iset_neg, iset_eq);
      // Add it to nsstar_nonstrict too.
      // FIXME: move?
      add_to_minimal_set_of_ns(ns_star, nsstar_nonstrict);
    }
  }

  // Add to nsstar_strict new ns from EQ.
  create_ns_from_eq_points(sk_strict, ns_strict, sk_ranges, ns_ranges_strict,
                           nsstar_strict, iset_pos, iset_eq, iset_neg,
                           sat_c, sat_g, min_sat, max_sat);

  if (should_create_nonstrict) {
    // Add to nsstar_nonstrict new ns from POS.
    create_ns_from_Q_with_face_enum<Range::POS>(Range::NEG,
                                                sk_nonstrict, ns_nonstrict,
                                                sk_ranges, ns_ranges_nonstrict,
                                                nsstar_nonstrict,
                                                iset_pos, iset_eq, iset_neg,
                                                sat_c, sat_g,
                                                false);
  }
}

void
split_non_skel(NS_Rows& ns_strict, NS_Rows& ns_nonstrict,
               Gens& sk_strict, Gens& sk_nonstrict,
               const Ranges& sk_ranges,
               const Sat& sat_c, Sat& sat_g,
               dim_type space_dim, dim_type num_lines, dim_type num_eq) {
  NS_Rows nsstar_strict, nsstar_nonstrict;
  Ranges ns_ranges_strict, ns_ranges_nonstrict;
  // Create partition for non-skeleton rows.
  Index_Set iset_neg, iset_eq, iset_pos;
  sk_ranges_to_index_sets(sk_ranges, iset_neg, iset_eq, iset_pos);
  ns_partition(ns_strict, iset_neg, iset_pos, ns_ranges_strict);

  // Copy partition, since they will follow different ns elements.
  // NB: the real nonstrict partition should be its opposite,
  // it is not complemented for efficiency.
  ns_nonstrict = ns_strict;
  ns_ranges_nonstrict = ns_ranges_strict;
  nsstar_strict.clear();
  nsstar_nonstrict.clear();

  bool sat_is_up_to_date = false;
  const Index_Partition<Range::POS_NEG> ns_pos_neg(ns_ranges_strict);
  if (!ns_pos_neg.empty()) {
    sat_g.transpose_assign(sat_c);
    sat_is_up_to_date = true;
    split_move_ns(ns_strict, ns_nonstrict,
                  ns_ranges_strict, ns_ranges_nonstrict,
                  sk_nonstrict,
                  iset_neg, iset_eq, iset_pos,
                  nsstar_strict, nsstar_nonstrict,
                  sat_c, sat_g);
  }

  // Check if we need to call `create_ns' for the non-strict side too.
  const bool should_create_nonstrict =
    can_have_nonskel(iset_eq, sk_nonstrict);
  if (!sat_is_up_to_date) {
    sat_g.transpose_assign(sat_c);
    sat_is_up_to_date = true;
  }
  const dim_type min_sat = space_dim - num_eq - 1;
  const dim_type max_sat = sat_c.num_cols() + num_lines;
  split_create_ns(ns_strict, ns_nonstrict,
                  ns_ranges_strict, ns_ranges_nonstrict,
                  sk_strict, sk_nonstrict, sk_ranges,
                  iset_pos, iset_eq, iset_neg,
                  nsstar_strict, nsstar_nonstrict,
                  sat_c, sat_g, min_sat, max_sat,
                  should_create_nonstrict);
  // Add new ns.
  if (sat_is_up_to_date) {
    new_ns_from_set(sk_strict, ns_strict, nsstar_strict,
                    sk_ranges, ns_ranges_strict, true);
    new_ns_from_set(sk_nonstrict, ns_nonstrict, nsstar_nonstrict,
                    sk_ranges, ns_ranges_nonstrict, false);
  }

  // Final update
  std::replace(ns_ranges_strict.begin(), ns_ranges_strict.end(),
               Range::POS_NEG, Range::NEG);
  std::replace(ns_ranges_strict.begin(), ns_ranges_strict.end(),
               Range::EQ, Range::NEG);
  std::replace(ns_ranges_nonstrict.begin(), ns_ranges_nonstrict.end(),
               Range::POS_NEG, Range::POS);
  Index_Partition<Range::NEG> ns_neg(ns_ranges_strict);
  Index_Partition<Range::POS> ns_pos(ns_ranges_nonstrict);
  erase_using_sorted_indices(ns_strict, ns_neg);
  erase_using_sorted_indices(ns_nonstrict, ns_pos);
}

/*******************************************************/
// Non-skel handling in CLOSED split.

/* Overloaded move-ns method for CLOSED split */
// Compute the closures and project in EQ.
// Check the supports to be promoted.
void
split_move_ns(NS_Rows& ns_rows, Ranges& ns_ranges, Gens& sk_rows,
              const Index_Set& iset_neg,
              const Index_Set& iset_eq,
              NS_Rows& nsstar,
              const Sat& sat_c, Sat& sat_g) {
  // Originally ns_ranges are the same.
  Index_Partition<Range::POS_NEG> ns_pos_neg(ns_ranges);
  Dims to_be_del;
  for (auto i : ns_pos_neg) {
    assert(ns_ranges[i] == Range::POS_NEG);
    Index_Set& supp = ns_rows[i];
    support_closure(supp, sat_c, sat_g);
    supp_projection(supp, false, iset_neg, iset_eq);

    // Check if it can be moved into SK.
    if (supp.size() == 1) {
      // Change its only supporter's type into point.
      const auto supporter_index = *(supp.begin());
      promote_singleton(sk_rows[supporter_index]);
      to_be_del.push_back(i);
      // Consider next ns in pos_neg
      continue;
    }
    add_to_minimal_set_of_ns(supp, nsstar);
  }
  // Remove ns that has been moved into SK.
  erase_using_sorted_indices(ns_rows, to_be_del);
  erase_using_sorted_indices(ns_ranges, to_be_del);
}

/** Overloaded create_ns method for CLOSED split. **/
// Add to NS*:
// - ns* from NEG projected in EQ,
// - ns* from POS projected in EQ.
void
split_create_ns(NS_Rows& ns_rows, Ranges& ns_ranges,
                Gens& sk_rows, const Ranges& sk_ranges,
                const Index_Set& iset_pos,
                const Index_Set& iset_eq,
                const Index_Set& iset_neg,
                NS_Rows& nsstar,
                const Sat& sat_c, const Sat& sat_g) {
  create_ns_from_Q_with_face_enum<Range::NEG>(Range::POS,
                                              sk_rows, ns_rows,
                                              sk_ranges, ns_ranges,
                                              nsstar,
                                              iset_pos, iset_eq, iset_neg,
                                              sat_c, sat_g,
                                              false);
  create_ns_from_Q_with_face_enum<Range::POS>(Range::NEG,
                                              sk_rows, ns_rows,
                                              sk_ranges, ns_ranges,
                                              nsstar,
                                              iset_pos, iset_eq, iset_neg,
                                              sat_c, sat_g,
                                              false);
}

/* Overloaded version for CLOSED split */
// Computes only one NS* and copies the whole NS sys only in the end.
void
split_non_skel(NS_Rows& ns_strict, NS_Rows& ns_nonstrict,
               Gens& sk_rows, const Ranges& sk_ranges,
               const Sat& sat_c, Sat& sat_g) {
  NS_Rows nsstar_strict;
  Ranges ns_ranges_strict;
  // Create partition for non-skeleton rows.
  Index_Set iset_neg, iset_eq, iset_pos;
  sk_ranges_to_index_sets(sk_ranges, iset_neg, iset_eq, iset_pos);
  ns_partition(ns_strict, iset_neg, iset_pos, ns_ranges_strict);
  nsstar_strict.clear();

  bool sat_is_up_to_date = false;
  const Index_Partition<Range::POS_NEG> ns_pos_neg(ns_ranges_strict);
  if (!ns_pos_neg.empty()) {
    sat_g.transpose_assign(sat_c);
    sat_is_up_to_date = true;
    split_move_ns(ns_strict, ns_ranges_strict, sk_rows,
                  iset_neg, iset_eq,
                  nsstar_strict, sat_c, sat_g);
  }

  // Check if we need to call `create_ns'.
  const bool should_create_nonstrict =
    can_have_nonskel(iset_eq, sk_rows);
  if (should_create_nonstrict) {
    if (!sat_is_up_to_date) {
      sat_g.transpose_assign(sat_c);
      sat_is_up_to_date = true;
    }
    split_create_ns(ns_strict, ns_ranges_strict,
                    sk_rows, sk_ranges,
                    iset_pos, iset_eq, iset_neg,
                    nsstar_strict,
                    sat_c, sat_g);
  }
  // Add new ns.
  if (sat_is_up_to_date) {
    new_ns_from_set(sk_rows, ns_strict, nsstar_strict,
                    sk_ranges, ns_ranges_strict, true);
  }

  // Copy NS systems.
  Ranges ns_ranges_nonstrict = ns_ranges_strict;
  ns_nonstrict = ns_strict;

  // Final update
  std::replace(ns_ranges_strict.begin(), ns_ranges_strict.end(),
               Range::POS_NEG, Range::NEG);
  std::replace(ns_ranges_nonstrict.begin(), ns_ranges_nonstrict.end(),
               Range::POS_NEG, Range::POS);
  Index_Partition<Range::NEG> ns_neg(ns_ranges_strict);
  Index_Partition<Range::POS> ns_pos(ns_ranges_nonstrict);
  erase_using_sorted_indices(ns_strict, ns_neg);
  erase_using_sorted_indices(ns_nonstrict, ns_pos);
}

/*******************************************************/
// Process violating line

void
split_violating_singular(dim_type ex,
                         Gen_Sys& sys_strict,
                         Gen_Sys& sys_nonstrict,
                         Sat& sat_c_strict, Sat& sat_c_nonstrict,
                         const Con& beta_strict) {
  auto& sg_strict = sys_strict.sing_rows;
  auto& sk_strict = sys_strict.sk_rows;
  auto& ns_strict = sys_strict.ns_rows;
  auto& sk_nonstrict = sys_nonstrict.sk_rows;

  const bool nnc_split = beta_strict.is_strict_inequality();

  // Map the ex_strict in a ray
  auto& ex_strict = sg_strict[ex];
  Integer sp_ex;
  Scalar_Prod::assign(sp_ex, beta_strict, ex_strict);
  make_non_singular(ex_strict, sp_ex);
  // Map all other geometric rows into SK=: can be done once.
  combine_with_ex_sing(ex, sp_ex, beta_strict, sg_strict, sk_strict);
  // Move the ex singular to sk
  const auto new_ex = num_rows(sk_strict);
  sk_strict.push_back(std::move(ex_strict));
  sg_strict.erase(sg_strict.begin() + ex);
  // Copy sys: change ray orientation for the non-strict side.
  sys_nonstrict = sys_strict;
  auto& ex_nonstrict = sk_nonstrict[num_rows(sk_nonstrict) - 1];
  assert(ex_nonstrict == sk_strict[new_ex]);
  neg_assign(ex_nonstrict.linear_expr());

  // Update sat_c : can be done once.
  assert(beta_strict.is_inequality());
  sat_c_strict.add_cols(1);
  auto sat_ex = Bit_Row();
  sat_ex.set(sat_c_strict.num_cols() - 1);
  sat_c_strict.add_row(std::move(sat_ex));
  // Copy sat: sat info are shared.
  sat_c_nonstrict = sat_c_strict;

  // Here sk_nonstrict and ns_nonstrict are done.

  if (nnc_split) {
    // Handle ns_strict and sk_strict eq points.
    Ranges sk_ranges, ns_ranges;
    sk_ranges.assign(sk_strict.size(), Range::EQ);
    sk_ranges[new_ex] = Range::POS;
    ns_ranges.assign(ns_strict.size(), Range::EQ);
    create_ns_with_ex_sing(sk_strict, ns_strict,
                           sk_ranges, ns_ranges, new_ex);
  }
}

/********************************************************/

/**
 * Split with RATIONAL VARIABLES.
 * This is the usual conversion algorithm, only
 * updating gen_sys_then and gen_sys_else differently.
 */
template <bool rational>
Empty_Pair_Res
split_gen_sys(dim_type space_dim, dim_type num_eq,
              Gen_Sys& gen_sys_strict, Gen_Sys& gen_sys_nonstrict,
              Sat& sat_c_strict, Sat& sat_c_nonstrict,
              Sat& sat_g_strict, Sat& sat_g_nonstrict,
              const Con& beta_strict,
              /* used only with non-rational vars*/
              const Con& beta_nonstrict,
              bool nnc_poly) {
  // Avoid warnings.
  (void) beta_nonstrict;
  // If `nnc_split' then the "strict" side is actually strict.
  const bool nnc_split = beta_strict.is_strict_inequality();
  assert(!nnc_split || nnc_poly);

  Gens& sg_strict = gen_sys_strict.sing_rows;
  Gens& sk_strict = gen_sys_strict.sk_rows;
  Gens& sk_nonstrict = gen_sys_nonstrict.sk_rows;
  NS_Rows& ns_strict = gen_sys_strict.ns_rows;
  NS_Rows& ns_nonstrict = gen_sys_nonstrict.ns_rows;

  // Check if all singular rows saturate strict_beta.
  const dim_type ex = check_sing_rows(beta_strict, sg_strict);
  if (ex < num_rows(sg_strict)) {
    split_violating_singular(ex, gen_sys_strict,
                             gen_sys_nonstrict,
                             sat_c_strict, sat_c_nonstrict,
                             beta_strict);
    return std::make_pair(false, false);
  }

  // Here all lines saturate the constraint.
  // Copy sing sys
  gen_sys_nonstrict.sing_rows = sg_strict;
  dim_type num_lines = num_rows(sg_strict);

  // Split the skeleton part.
  // Copies the sk-sys only if nnc_split and if both sides are non-empty.
  // Updates only `sat_strict': the copy is posponed.
  Ranges sk_ranges;
  auto res = split_skel(space_dim, num_lines, num_eq,
                        sk_strict, sk_nonstrict,
                        sk_ranges, beta_strict,
                        ns_strict,
                        sat_c_strict);

  if (res.first || res.second) {
    return res;
  }

  // Non-skel handling is required only in NNC polyhedra.
  if (nnc_poly) {
    if (nnc_split) {
      split_non_skel(ns_strict, ns_nonstrict,
                     sk_strict, sk_nonstrict, sk_ranges,
                     sat_c_strict, sat_g_strict,
                     space_dim, num_lines, num_eq);
    } else {
      split_non_skel(ns_strict, ns_nonstrict, sk_strict, sk_ranges,
                     sat_c_strict, sat_g_strict);
    }
  }

  if (!res.second) {
    // Copy sat: all saturation info are shared.
    sat_c_nonstrict = sat_c_strict;
    // ANNA: FIXME Since simplify will be called next, is this needed?
    sat_g_nonstrict = sat_g_strict;
    if (!nnc_split)
      // Copy SK sys since it has been computed only once.
      sk_nonstrict = sk_strict;
  }

  // Final SK update.
  Index_Partition<Range::NEG> sk_neg(sk_ranges);
  Index_Partition<Range::EQ> sk_eq(sk_ranges);
  Index_Partition<Range::POS> sk_pos(sk_ranges);
  if (nnc_split)
    points_become_closure_points(sk_eq, sk_strict);
  remove_rows(sk_neg, sk_strict, ns_strict, sat_c_strict);
  remove_rows(sk_pos, sk_nonstrict, ns_nonstrict, sat_c_nonstrict);

  return res;
}

/***********************************/

// Specialization for NON-RATIONAL VARIABLES.
// The whole impl should be different...?
template<>
Empty_Pair_Res
split_gen_sys<false>(dim_type space_dim, dim_type num_eq,
                     Gen_Sys& gen_sys_then, Gen_Sys& gen_sys_else,
                     Sat& sat_c_then, Sat& sat_c_else,
                     Sat& sat_g_then, Sat& sat_g_else,
                     const Con& beta_strict,
                     /* used only with non-rational vars*/
                     const Con& beta_nonstrict,
                     bool nnc_poly);

//////////////////////////////////////////////////////////////

/**
 * Split method.
 * Updates ph_then = ph_then.add_con(beta)
 *         ph_else = ph_then.add_con(beta_complement)
 */
template <bool rational>
void
split_poly(Poly_Impl& ph_then, Poly_Impl& ph_else,
           const Con& beta, Topol t) {
  assert(ph_then.is_minimized());
  const bool nnc_poly = (ph_then.topology() == Topol::NNC);
  assert(t == Topol::CLOSED || nnc_poly);

  if (ph_then.is_empty()) {
    ph_else = ph_then;
    return;
  }

  if (ph_then.space_dim() == 0) {
    ph_else = ph_then;
    if (beta.is_inconsistent())
      ph_then.set_empty();
    return;
  }

  ph_else.set_status(Poly_Impl::Status::MINIMIZED);

  const dim_type space_dim = ph_then.space_dim();
  // FIXME: when to update this???
  dim_type num_eq = num_rows(ph_then.cs.sing_rows);

  assert(ph_else.gs.empty());
  // Copy original constraint system
  ph_else.cs = ph_then.cs;

  Con beta_else = complement_con(beta, t);
  const bool strict = beta.is_strict_inequality();
  const bool then_is_strict = (strict || t == Topol::CLOSED);

  // [non]strict suffixes are "real" only if t == Topol::NNC.
  auto& ph_strict = then_is_strict ? ph_then : ph_else;
  auto& ph_nonstrict = then_is_strict ? ph_else : ph_then;
  const auto& beta_strict = then_is_strict ? beta : beta_else;
  const auto& beta_nonstrict = then_is_strict ? beta_else : beta;
  auto& gens_strict = ph_strict.gs;
  auto& gens_nonstrict = ph_nonstrict.gs;
  auto& sat_c_strict = ph_strict.sat_c;
  auto& sat_c_nonstrict = ph_nonstrict.sat_c;
  auto& sat_g_strict = ph_strict.sat_g;
  auto& sat_g_nonstrict = ph_nonstrict.sat_g;
  auto& cons_strict = ph_strict.cs;
  auto& cons_nonstrict = ph_nonstrict.cs;

  // The "strict" side holds the original system.
  if (!then_is_strict) {
    // FIXME: this may be heavy...
    std::swap(gens_strict, gens_nonstrict);
    std::swap(sat_c_strict, sat_c_nonstrict);
    std::swap(sat_g_strict, sat_g_nonstrict);
  }

  assert((t == Topol::NNC) == beta_strict.is_strict_inequality());
  const Empty_Pair_Res empty_pair =
    split_gen_sys<rational>(space_dim, num_eq,
                            gens_strict, gens_nonstrict,
                            sat_c_strict, sat_c_nonstrict,
                            sat_g_strict, sat_g_nonstrict,
                            beta_strict, beta_nonstrict,
                            nnc_poly);

  // If both results are non-empty, then `beta' is not redundant.
  // Update cons systems.
  if (!empty_pair.first && !empty_pair.second) {
    cons_nonstrict.add(beta_nonstrict);
    cons_strict.add(beta_strict);
  }

  // If the strict side is empty, `beta' is redundant in the non-strict side.
  // No need to update cons sys; recover ph_nonstrict as the original ph.
  if (empty_pair.first && !empty_pair.second) {
    std::swap(ph_strict, ph_nonstrict);
  }

  if (empty_pair.first)
    ph_strict.set_empty();
  else
    simplify(space_dim, cons_strict, sat_c_strict, sat_g_strict, gens_strict);

  if (empty_pair.second)
    ph_nonstrict.set_empty();
  else
    simplify(space_dim, cons_nonstrict, sat_c_nonstrict, sat_g_nonstrict, gens_nonstrict);

}

template void
split_poly<true>(Poly_Impl& ph_then, Poly_Impl& ph_else,
                 const Con& beta, Topol t);

} // namespace detail
} // namespace pplite
